"""
This module contains all of the information for the titratable residues,
including reference energies, model compound pKas, and charge vectors for every
titratable residue treated.
"""
from __future__ import print_function, division

titratable_residues = ['AS4', 'GL4', 'CYS', 'TYR', 'HIP', 'LYS', 'DAP', 'DCP',
                       'DG', 'DT', 'AP', 'CP', 'G', 'U', 'HEH', 'PRN', 'TYX']

from math import log
from io import StringIO
from ..exceptions import AmberWarning, AmberError
from ..utils.six.moves import range
import warnings

# Print all AmberWarning's
warnings.filterwarnings('always', category=AmberWarning)

class _State(object):
    """ A protonation state """
    sort_by_resnum = True

    def __init__(self, charges, refene, refene_old=None, protcnt=None, pka_corr=None, eleccnt=None, eo_corr=None):
        self.charges = charges
        self.refene = refene
        self.refene_old = refene_old
        self.protcnt = protcnt
        self.pka_corr = pka_corr
        self.eleccnt = eleccnt
        self.eo_corr = eo_corr

class _ReferenceEnergy(object):
    """ Reference energies for various solvent models """
    LN_TO_LOG = log(10.0)
    KB = 0.00199
    TEMP = 300.0
    def __init__(self, igb1=None, igb2=None, igb5=None, igb7=None, igb8=None):
        self.pKa_is_set = False
        self.igb1 = igb1
        self.igb2 = igb2
        self.igb5 = igb5
        self.igb7 = igb7
        self.igb8 = igb8

    def solvent_energies(self, igb1=None, igb2=None, igb5=None,
                         igb7=None, igb8=None):
        """
        Add solvent reference energies, copying the GB reference energies if
        none are explicitly given
        """
        self.solvent = _ReferenceEnergy(igb1, igb2, igb5, igb7, igb8)

    def dielc2_energies(self, igb1=None, igb2=None, igb5=None,
                        igb7=None, igb8=None):
        """
        Add reference energies for a dielectric constant of 2.0. Since this has
        no reason to be near the original reference energies, do not use those
        in place of energies that aren't provided
        """
        self.dielc2 = _ReferenceEnergy(igb1, igb2, igb5, igb7, igb8)

    def set_pKa(self, pKa, deprotonated=False):
        """
        Adjusts the reference energies based on the pKa. If the reference energy
        is given for a deprotonated state, the pKa adjustment is subtracted from
        the reference energy. Otherwise, this state is a 'protonated' state, so
        the pKa adjustment is added to the reference energy
        """
        if self.pKa_is_set: return
        self.pKa_is_set = True
        factor = self.KB * self.LN_TO_LOG * self.TEMP * pKa

        if deprotonated:
            if self.igb1 is not None: self.igb1 -= factor
            if self.igb2 is not None: self.igb2 -= factor
            if self.igb5 is not None: self.igb5 -= factor
            if self.igb7 is not None: self.igb7 -= factor
            if self.igb8 is not None: self.igb8 -= factor
        else:
            if self.igb1 is not None: self.igb1 += factor
            if self.igb2 is not None: self.igb2 += factor
            if self.igb5 is not None: self.igb5 += factor
            if self.igb7 is not None: self.igb7 += factor
            if self.igb8 is not None: self.igb8 += factor
        # Now apply the correction to our solvent reference energies if we have
        # them (and dielectric-2 energies, if we have them)
        if hasattr(self, 'solvent'):
            self.solvent.set_pKa(pKa, deprotonated)
        if hasattr(self, 'dielc2'):
            self.dielc2.set_pKa(pKa, deprotonated)

class _LineBuffer(object):
    """ Buffer to add lines to the cpin file """

    CHARS_PER_LINE = 80

    def __init__(self, file):
        self.file = file
        self.linebuffer = ''

    def add_word(self, word):
        if len(self.linebuffer) + len(word) > self.CHARS_PER_LINE:
            self.file.write(self.linebuffer + '\n')
            self.linebuffer = ' %s' % word
        else:
            self.linebuffer += word

    def add_words(self, words, space_delimited=False):
        """ Adds multiple words """
        extra = ''
        if space_delimited:
            extra = ' '
        for word in words:
            self.add_word(word + extra)

    def flush(self):
        """ Flushes this buffer to the file """
        if len(self.linebuffer) == 0:
            return
        self.file.write(self.linebuffer + '\n')
        self.linebuffer = ''

class TitratableResidue(object):
    """
    A residue with different protonation states defined for Amber for use in
    the Constant pH MD method implemented in sander
    """

    def __init__(self, resname, atom_list, typ, pka=None, eo=None):
        self.resname = resname
        self.atom_list = list(atom_list) # list of atom names
        self.states = []
        self.first_state = -1
        self.first_charge = -1
        self.typ = typ
        self.pKa = pka
        self.Eo = eo

    def _str_refenes(self, solvent=False, igb=2, dielc=1.0):
        """
        Converts all reference energies into a formatted string with a message
        saying if the energy is not set
        """
        ret_str = ''
        for state in self.states:
            if dielc == 2:
                if solvent:
                    refene = getattr(state.refene.dielc2.solvent, 'igb%d'%igb)
                else:
                    refene = getattr(state.refene.dielc2, 'igb%d'%igb)
            else:
                if solvent:
                    refene = getattr(state.refene.solvent, 'igb%d'%igb)
                else:
                    refene = getattr(state.refene, 'igb%d'%igb)

            if refene is None:
                ret_str += '%12s' % 'Not Set'
            else:
                ret_str += '%12.5f' % refene

        return ret_str

    def __str__(self):
        if (self.typ=="ph"):
            ret_str = ('%-4s\tpKa = %5.1f\n%8s' % (self.resname, self.pKa, 'ATOM') +
                       ''.join(['%12s' % ('STATE %d' % i) for i in
                               range(len(self.states))]) + '\n'
            )
        elif (self.typ=="redox"):
            ret_str = ('%-4s\tEo = %7.3f V\n%8s' % (self.resname, self.Eo, 'ATOM') +
                       ''.join(['%12s' % ('STATE %d' % i) for i in
                               range(len(self.states))]) + '\n'
            )
        else:
            ret_str = ('%-4s\n%8s' % (self.resname, 'ATOM') +
                       ''.join(['%12s' % ('STATE %d' % i) for i in
                               range(len(self.states))]) + '\n'
            )
        for i, atom in enumerate(self.atom_list):
            ret_str += ('%8s' % atom +
                        ''.join(['%12.4f' % (state.charges[i])
                                for state in self.states]) + '\n'
            )
        ret_str += '-' * (8 + 12 * len(self.states)) + '\n'
        if (self.typ == "ph" or self.typ == "phredox"):
            ret_str += ('%8s' % 'Prot Cnt' +
                        ''.join(['%12d' % state.protcnt for state in self.states]) +
                        '\n')
            ret_str += '-' * (8 + 12 * len(self.states)) + '\n'
            ret_str += ('%8s' % 'pKa Corr' +
                        ''.join(['%12.4f' % state.pka_corr for state in self.states]) +
                        '\n')
        if (self.typ == "phredox"):
            ret_str += '-' * (8 + 12 * len(self.states)) + '\n'
        if (self.typ == "redox" or self.typ == "phredox"):
            ret_str += ('%8s' % 'Elec Cnt' +
                        ''.join(['%12d' % state.eleccnt for state in self.states]) +
                        '\n')
            ret_str += '-' * (8 + 12 * len(self.states)) + '\n'
            ret_str += ('%8s' % 'Eo Corr' +
                        ''.join(['%12.4f' % state.eo_corr for state in self.states]) +
                        '\n')
        ret_str += '-' * (8 + 12 * len(self.states)) + '\n'
        ret_str += ('Reference Energies (ES = Explicit solvent, IS = Implicit '
                    'solvent)\n\n')
        ret_str += '%8s' % ('igb=1 IS') + self._str_refenes(False, 1) + '\n'
        ret_str += '%8s' % ('igb=2 IS') + self._str_refenes(False, 2) + '\n'
        ret_str += '%8s' % ('igb=5 IS') + self._str_refenes(False, 5) + '\n'
        ret_str += '%8s' % ('igb=7 IS') + self._str_refenes(False, 7) + '\n'
        ret_str += '%8s' % ('igb=8 IS') + self._str_refenes(False, 8) + '\n'
        ret_str += '%8s' % ('igb=1 ES') + self._str_refenes(True, 1) + '\n'
        ret_str += '%8s' % ('igb=2 ES') + self._str_refenes(True, 2) + '\n'
        ret_str += '%8s' % ('igb=5 ES') + self._str_refenes(True, 5) + '\n'
        ret_str += '%8s' % ('igb=7 ES') + self._str_refenes(True, 7) + '\n'
        ret_str += '%8s' % ('igb=8 ES') + self._str_refenes(True, 8) + '\n'
        ret_str += '-' * (8 + 12 * len(self.states)) + '\n'
        ret_str += 'Reference Energies for Internal Dielectric of 2.0\n\n'
        ret_str += '%8s' % ('igb=1 IS') + self._str_refenes(False, 1, 2) + '\n'
        ret_str += '%8s' % ('igb=2 IS') + self._str_refenes(False, 2, 2) + '\n'
        ret_str += '%8s' % ('igb=5 IS') + self._str_refenes(False, 5, 2) + '\n'
        ret_str += '%8s' % ('igb=7 IS') + self._str_refenes(False, 7, 2) + '\n'
        ret_str += '%8s' % ('igb=8 IS') + self._str_refenes(False, 8, 2) + '\n'
        ret_str += '%8s' % ('igb=1 ES') + self._str_refenes(True, 1, 2) + '\n'
        ret_str += '%8s' % ('igb=2 ES') + self._str_refenes(True, 2, 2) + '\n'
        ret_str += '%8s' % ('igb=5 ES') + self._str_refenes(True, 5, 2) + '\n'
        ret_str += '%8s' % ('igb=7 ES') + self._str_refenes(True, 7, 2) + '\n'
        ret_str += '%8s' % ('igb=8 ES') + self._str_refenes(True, 8, 2) + '\n'
        return ret_str

    def add_state(self, charges, refene, refene_old=None, protcnt=None, pka_corr=None, eleccnt=None, eo_corr=None):
        """ Add a single titratable state for this titratable residue """
        new_state = _State(charges, refene, refene_old=refene_old, protcnt=protcnt, pka_corr=pka_corr, eleccnt=eleccnt, eo_corr=eo_corr)
        if len(new_state.charges) != len(self.atom_list):
            raise AmberError('Wrong number of charges for new state')
        self.states.append(new_state)

    def add_states(self, charges, refenes, refenes_old=None, protcnts=None, pka_corrs=None, eleccnts=None, eo_corrs=None):
        """ Add multiple titratable states for this titratable residue """
        if len(charges) != len(refenes) or (len(charges) != len(refenes_old) and refenes_old) or (len(charges) != len(protcnts) and protcnts) \
           or (len(charges) != len(pka_corrs) and pka_corrs) or (len(charges) != len(eleccnts) and eleccnts) \
           or (len(charges) != len(eo_corrs) and eo_corrs):
            raise AmberError('Inconsistent list of parameters for '
                             'TitratableResidue.add_states')
        for i in range(len(charges)):
            self.add_state(charges[i], refenes[i], refenes_old[i], protcnts[i], pka_corrs[i], eleccnts[i], eo_corrs[i])

    def cpin_pointers(self, first_atom):
        """ Sets and returns the cpin info """
        if self.first_state == -1 or self.first_charge == -1:
            raise AmberError('Must set residue pointers before writing '
                             'cpin info!')
        return {'FIRST_ATOM' : first_atom,
                'FIRST_CHARGE' : self.first_charge,
                'FIRST_STATE' : self.first_state,
                'NUM_ATOMS' : len(self.atom_list),
                'NUM_STATES' : len(self.states)}

    def set_first_state(self, index):
        """ Sets the first state index """
        # Has the first state already been set?
        if self.first_state != -1:
            if index != self.first_state:
                raise AmberError('First state already set differently')
        self.first_state = index

    def set_first_charge(self, index):
        """ Sets the first charge index """
        # Has it already been set?
        if self.first_charge != -1:
            if index != self.first_charge:
                raise AmberError('First charge already set differently')
        self.first_charge = index

    def reset(self):
        """ Resets the pointers """
        self.first_state = -1
        self.first_charge = -1

    def check(self):
        """ Checks that the charges are consistent w/ the protonation states """
        sum_charges = [sum(state.charges) for state in self.states]
        if (self.typ == "ph" or self.typ == "phredox"):
            protcnts = [state.protcnt for state in self.states]
        if (self.typ == "redox" or self.typ == "phredox"):
            eleccnts = [state.eleccnt for state in self.states]
        # All we have to do is make sure that the charges/proton counts are
        # consistent between the first state and every other state
        for i in range(1, len(sum_charges)):
            charge_diff = sum_charges[i] - sum_charges[0]
            if (self.typ == "ph"):
                diff = protcnts[i] - protcnts[0]
            elif (self.typ == "redox"):
                diff = eleccnts[0] - eleccnts[i]
            elif (self.typ == "phredox"):
                diff = protcnts[i] - protcnts[0] + eleccnts[0] - eleccnts[i]
            if abs(charge_diff - diff) >= 0.0001:
                warnings.warn('Inconsistencies detected in charge definitions '
                              'in %s' % self.resname, AmberWarning)
        # Check all of the reference energies to make sure that the pKa was set
        # for all but one of them
        notset = 0
        valid = True
        for state in self.states:
            if (not state.refene_old):
                valid = False
                break
            notset += int(state.refene_old.pKa_is_set)
        if (notset != len(self.states) - 1 and valid):
            warnings.warn('Not enough states are pKa-adjusted in %s' %
                          self.resname)

class TitratableResidueList(list):
    """ List of all titratable residues """
    def __init__(self, system_name='Unknown', solvated=False,
                 first_solvent=0):
        list.__init__(self)
        self.first_atoms = []
        self.residue_nums = []
        self.resstates = []
        self.system_name = system_name
        self.solvated = solvated
        self.first_sol = first_solvent

    def add_residue(self, residue, resnum, first_atom, state=0):
        """ Adds a residue to the list """
        list.append(self, residue)
        self.first_atoms.append(first_atom)
        self.residue_nums.append(resnum)
        if state < 0 or state >= len(residue.states):
            raise AmberError('Residue %s only has states 0-%d (%d chosen)' %
                        (residue.resname, len(residue.states)-1, state))
        self.resstates.append(state)

    def set_states(self, statelist):
        """
        Sets the initial protonation states from a list -- make sure there are
        enough states in the list to set every residue, or emit a warning
        """
        if len(statelist) != len(self):
            warnings.warn(('Number of states (%d) does not equal number of '
                            'residues (%d). Using default initial states.') %
                            (len(statelist), len(self)), AmberWarning)
            return
        # Check that all states are allowable
        for i, state in enumerate(statelist):
            if state < 0 or state >= len(self[i].states):
                raise AmberError('Bad state choice (%d). Minimum is 0, '
                            'maximum is %d' % (state, len(self[i].states)))
        # If we got here, then we are OK
        self.resstates = statelist

    def sort(self):
        """ Sorts by residue number """
        # Bubble sort, cuz who cares?
        nswaps = 1
        while nswaps > 0:
            nswaps = 0
            for i in range(len(self)-1):
                if self.first_atoms[i] > self.first_atoms[i+1]:
                    nswaps += 1
                    self.first_atoms[i], self.first_atoms[i+1] = \
                        self.first_atoms[i+1], self.first_atoms[i]
                    self[i], self[i+1] = self[i+1], self[i]
                    self.residue_nums[i], self.residue_nums[i+1] = \
                        self.residue_nums[i+1], self.residue_nums[i]

    def write_cpin(self, output, igb=2, intdiel=1.0, oldfmt=False, typ="ph", coions=False):
        """ Writes the CPIN file based on the titrated residues """
        end = StringIO()
        # Reset all residues
        for res in self: res.reset()
        # Sort our residue list
        self.sort()
        limit_buf = _LineBuffer(output)
        buf = _LineBuffer(end)
        limit_buf.add_word('&CNSTPHE_LIMITS')
        limit_buf.flush()
        limit_buf.add_word(' ntres=%d,' % len(self))
        limit_buf.add_word(' maxh=%d,' % max(len(r.states) for r in self))
        if (typ == "ph"):
            buf.add_word('&CNSTPH')
        elif (typ == "redox"):
            buf.add_word('&CNSTE')
        elif (typ == "phredox"):
            buf.add_word('&CNSTPHE')
        buf.flush()
        buf.add_word(' CHRGDAT=')
        charges, energies, protcnts, pka_corrs, eleccnts, eo_corrs, pointers = [], [], [], [], [], [], []
        first_charge = 0
        first_state = 0
        for i, res in enumerate(self):
            if res.first_charge == -1:
                res.set_first_charge(first_charge)
                res.set_first_state(first_state)

                for state in res.states:
                    # See which dielectric reference energies we want
                    if intdiel == 2:
                        if (not oldfmt):
                            refene = state.refene.dielc2
                        else:
                            refene = state.refene_old.dielc2
                    else:
                        if (not oldfmt):
                            refene = state.refene
                        else:
                            refene = state.refene_old
                    # See if we want the explicit solvent refene or not
                    if self.solvated:
                        energies.append(getattr(refene.solvent, 'igb%d' % igb))
                    else:
                        energies.append(getattr(refene, 'igb%d' % igb))
                    if (typ == "ph" or typ == "phredox"):
                        # Add protonation count of this state
                        if (state.protcnt):
                            protcnts.append(state.protcnt)
                        else:
                            protcnts.append(0)
                        # Add pka reference of this state
                        if (state.pka_corr):
                            pka_corrs.append(state.pka_corr)
                        else:
                            pka_corrs.append(0.0)
                    if (typ == "redox" or typ == "phredox"):
                        # Add electron count of this state
                        if (state.eleccnt):
                            eleccnts.append(state.eleccnt)
                        else:
                            eleccnts.append(0)
                        # Add Eo reference of this state
                        if (state.eo_corr):
                            eo_corrs.append(state.eo_corr)
                        else:
                            eo_corrs.append(0.0)

                first_state += len(res.states)
                new_charges = []
                for state in res.states:
                    new_charges.extend(state.charges)
                charges.extend(new_charges)
                first_charge += len(new_charges)
            pointers.append(res.cpin_pointers(self.first_atoms[i]))

        limit_buf.add_word(' natchrg=%d,' % len(charges))
        limit_buf.add_word(' ntstates=%d' % max(len(protcnts), len(eleccnts)))
        limit_buf.flush()
        limit_buf.add_word('/')
        limit_buf.flush()
        # Print the charges
        for charge in charges:
            buf.add_word('%s,' % charge)
        buf.flush()
        # Print the protcnts
        if (typ == "ph" or typ == "phredox"):
            buf.add_word(' PROTCNT=')
            for protcnt in protcnts:
                buf.add_word('%d,' % protcnt)
        buf.flush()
        if (typ == "redox" or typ == "phredox"):
            buf.add_word(' ELECCNT=')
            for eleccnt in eleccnts:
                buf.add_word('%d,' % eleccnt)
        buf.flush()
        # Print the residue names
        buf.add_word(" RESNAME='System: %s'," % self.system_name)
        for i, res in enumerate(self):
            buf.add_word("'Residue: %s %d'," %
                    (res.resname, self.residue_nums[i]))
        buf.flush()
        # Print the residue states
        buf.add_word(" RESSTATE=")
        for state in self.resstates:
            buf.add_word('%d,' % state)
        buf.flush()
        # Print the residue pointers
        buf.add_word(' ') # get a leading space
        for i, p in enumerate(pointers):
            buf.add_word("STATEINF(%d)%%FIRST_ATOM=%d, " % (i, p['FIRST_ATOM']))
            buf.add_word("STATEINF(%d)%%FIRST_CHARGE=%d, " % (i, p['FIRST_CHARGE']))
            buf.add_word("STATEINF(%d)%%FIRST_STATE=%d, " % (i, p['FIRST_STATE']))
            buf.add_word("STATEINF(%d)%%NUM_ATOMS=%d, " % (i, p['NUM_ATOMS']))
            buf.add_word("STATEINF(%d)%%NUM_STATES=%d, " % (i, p['NUM_STATES']))
        buf.flush()
        # Print the reference energies
        buf.add_word(' STATENE=')
        for i, energy in enumerate(energies):
            if energy is None:
                raise AmberError("%d'th reference energy not known for igb = %d" % (i, igb))
            buf.add_word('%.6f,' % energy)
        buf.flush()
        # Print the pKa or Eo reference
        if (not oldfmt):
            if (typ == "ph" or typ == "phredox"):
                buf.add_word(' PKA_CORR=')
                for pka_corr in pka_corrs:
                    buf.add_word('%.4f,' % pka_corr)
            buf.flush()
            if (typ == "redox" or typ == "phredox"):
                buf.add_word(' EO_CORR=')
                for eo_corr in eo_corrs:
                    buf.add_word('%.4f,' % eo_corr)
            buf.flush()
        # Print the # of residues and explicit solvent info if required
        buf.add_word(' TRESCNT=%d,' % len(self))
        if self.solvated:
            if (typ == "ph"):
                buf.add_word('CPHFIRST_SOL=%d, CPH_IGB=%d, CPH_INTDIEL=%s, ' %
                            (self.first_sol, igb, intdiel))
            elif (typ == "redox"):
                buf.add_word('CEFIRST_SOL=%d, CE_IGB=%d, CE_INTDIEL=%s, ' %
                            (self.first_sol, igb, intdiel))
            elif (typ == "phredox"):
                buf.add_word('CPHEFIRST_SOL=%d, CPHE_IGB=%d, CPHE_INTDIEL=%s, ' %
                            (self.first_sol, igb, intdiel))
            buf.flush()
            # Now scan through all of the waters
        buf.flush()
        buf.add_word('/')
        buf.flush()
        end.seek(0)
        output.write(end.read())

# Now define all of the titratable residues

# Aspartate
refene1 = _ReferenceEnergy(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb1=21.4298008, igb2=26.8894581, igb5=26.5980488,
                 igb7=23.4181107, igb8=26.3448911)
refene2.solvent_energies(igb2=33.2613028, igb5=32.064349, igb7=28.262350, igb8=31.286037)
refene2.dielc2_energies(igb2=12.676908, igb5=13.084913)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb1=21.4298008, igb2=26.8894581, igb5=26.5980488,
                 igb7=23.4181107, igb8=26.3448911)
refene2_old.solvent_energies(igb2=33.2613028, igb5=32.064349, igb7=28.262350, igb8=31.286037)
refene2_old.dielc2_energies(igb2=12.676908, igb5=13.084913)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(4.0, deprotonated=False)

AS4 = TitratableResidue('AS4', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'OD1', 'OD2', 'HD21', 'C', 'O', 'HD22', 'HD11', 'HD12'],
                        pka=4.0, typ="ph")
AS4.add_state(protcnt=0, refene=refene1, refene_old=refene1, pka_corr=0.0, # deprotonated
              charges=[-0.4157, 0.2719, 0.0341, 0.0864, -0.1783, -0.0122,
              -0.0122, 0.7994, -0.8014, -0.8014, 0.0, 0.5973, -0.5679, 0.0, 0.0,
              0.0])
AS4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.0, # protonated syn-O2
              charges=[-0.4157, 0.2719, 0.0341, 0.0864, -0.0316, 0.0488, 0.0488,
              0.6462, -0.5554, -0.6376, 0.4747, 0.5973, -0.5679, 0.0, 0.0, 0.0])
AS4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.0, # protonated anti-O2
              charges=[-0.4157, 0.2719, 0.0341, 0.0864, -0.0316, 0.0488, 0.0488,
              0.6462, -0.5554, -0.6376, 0.0, 0.5973, -0.5679, 0.4747, 0.0, 0.0])
AS4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.0, # protonated syn-O1
              charges=[-0.4157, 0.2719, 0.0341, 0.0864, -0.0316, 0.0488, 0.0488,
              0.6462, -0.6376, -0.5554, 0.0, 0.5973, -0.5679, 0.0, 0.4747, 0.0])
AS4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.0, # protonated anti-O1
              charges=[-0.4157, 0.2719, 0.0341, 0.0864, -0.0316, 0.0488, 0.0488,
              0.6462, -0.6376, -0.5554, 0.0, 0.5973, -0.5679, 0.0, 0.0, 0.4747])
AS4.check() # check that everything is consistent

# Glutamate
refene1 = _ReferenceEnergy(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb1=3.89691326, igb2=8.4057785, igb5=8.0855764,
               igb7=5.305949, igb8=8.3591335)
refene2.solvent_energies(igb2=15.20019319, igb5=13.533409, igb7=10.682953, igb8=13.242410)
refene2.dielc2_energies(igb2=3.455596, igb5=3.957270)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb1=3.89691326, igb2=8.4057785, igb5=8.0855764,
               igb7=5.305949, igb8=8.3591335)
refene2_old.solvent_energies(igb2=15.20019319, igb5=13.533409, igb7=10.682953, igb8=13.242410)
refene2_old.dielc2_energies(igb2=3.455596, igb5=3.957270)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(4.4, deprotonated=False)

GL4 = TitratableResidue('GL4', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'HG2', 'HG3', 'CD', 'OE1', 'OE2', 'HE21', 'C', 'O',
                        'HE22', 'HE11', 'HE12'], pka=4.4, typ="ph")
GL4.add_state(protcnt=0, refene=refene1, refene_old=refene1, pka_corr=0.0, # deprotonated
              charges=[-0.4157, 0.2719, 0.0145, 0.0779, -0.0398, -0.0173,
              -0.0173, 0.0136, -0.0425, -0.0425, 0.8054, -0.8188, -0.8188, 0.0,
              0.5973, -0.5679, 0.0, 0.0, 0.0])
GL4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.4, # protonated syn-O2
              charges=[-0.4157, 0.2719, 0.0145, 0.0779, -0.0071, 0.0256, 0.0256,
              -0.0174, 0.0430, 0.0430, 0.6801, -0.5838, -0.6511, 0.4641, 0.5973,
              -0.5679, 0.0, 0.0, 0.0])
GL4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.4, # protonated anti-O2
              charges=[-0.4157, 0.2719, 0.0145, 0.0779, -0.0071, 0.0256, 0.0256,
              -0.0174, 0.0430, 0.0430, 0.6801, -0.5838, -0.6511, 0.0, 0.5973,
              -0.5679, 0.4641, 0.0, 0.0])
GL4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.4, # protonated syn-O1
              charges=[-0.4157, 0.2719, 0.0145, 0.0779, -0.0071, 0.0256, 0.0256,
              -0.0174, 0.0430, 0.0430, 0.6801, -0.6511, -0.5838, 0.0, 0.5973,
              -0.5679, 0.0, 0.4641, 0.0])
GL4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.4, # protonated syn-O2
              charges=[-0.4157, 0.2719, 0.0145, 0.0779, -0.0071, 0.0256, 0.0256,
              -0.0174, 0.0430, 0.0430, 0.6801, -0.6511, -0.5838, 0.0, 0.5973,
              -0.5679, 0.0, 0.0, 0.4641])
GL4.check()

# Tyrosine
refene1 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-65.113428, igb5=-64.166385, igb8=-61.3305355)
refene2.solvent_energies(igb2=-65.003415, igb5=-64.047229)
refene2.dielc2_energies(igb2=-32.167520, igb5=-31.751177)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=-65.113428, igb5=-64.166385, igb8=-61.3305355)
refene2_old.solvent_energies(igb2=-65.003415, igb5=-64.047229)
refene2_old.dielc2_energies(igb2=-32.167520, igb5=-31.751177)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(9.6, deprotonated=True)

TYR = TitratableResidue('TYR', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'CD1', 'HD1', 'CE1', 'HE1', 'CZ', 'OH', 'HH', 'CE2',
                        'HE2', 'CD2', 'HD2', 'C', 'O'], pka=9.6, typ="ph")
TYR.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=9.6, # protonated
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.0152, 0.0295,
              0.0295, -0.0011, -0.1906, 0.1699, -0.2341, 0.1656, 0.3226,
              -0.5579, 0.3992, -0.2341, 0.1656, -0.1906, 0.1699, 0.5973,
              -0.5679])
TYR.add_state(protcnt=0, refene=refene2, refene_old=refene2_old, pka_corr=0.0, # deprotonated
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.0858, 0.0190,
              0.0190, -0.2130, -0.1030, 0.1320, -0.4980, 0.1320, 0.7770,
              -0.8140, 0.0, -0.4980, 0.1320, -0.1030, 0.1320, 0.5973, -0.5679])
TYR.check()

# Histidine
refene1 = _ReferenceEnergy(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb1=-4.208863, igb2=-2.84183, igb5=-2.86001,
               igb7=-1.741947, igb8=-3.4000)
refene2.solvent_energies(igb2=-2.77641, igb5=-2.90517)
refene2.dielc2_energies(igb2=-1.628110, igb5=-1.691093)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb1=-4.208863, igb2=-2.84183, igb5=-2.86001,
               igb7=-1.741947, igb8=-3.4000)
refene2_old.solvent_energies(igb2=-2.77641, igb5=-2.90517)
refene2_old.dielc2_energies(igb2=-1.628110, igb5=-1.691093)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(6.5, deprotonated=True)
refene3 = _ReferenceEnergy(igb1=-8.230643, igb2=-6.58793, igb5=-6.70726,
               igb7=-5.118453, igb8=-6.3190)
refene3.solvent_energies(igb2=-6.483630, igb5=-6.82684)
refene3.dielc2_energies(igb2=-3.444200, igb5=-3.070113)
refene3.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene3_old = _ReferenceEnergy(igb1=-8.230643, igb2=-6.58793, igb5=-6.70726,
               igb7=-5.118453, igb8=-6.3190)
refene3_old.solvent_energies(igb2=-6.483630, igb5=-6.82684)
refene3_old.dielc2_energies(igb2=-3.444200, igb5=-3.070113)
refene3_old.dielc2.solvent_energies()
refene3_old.set_pKa(7.1, deprotonated=True)

HIP = TitratableResidue('HIP', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'ND1', 'HD1', 'CE1', 'HE1', 'NE2', 'HE2', 'CD2', 'HD2',
                        'C', 'O'], pka=6.6, typ="ph")
HIP.add_state(protcnt=2, refene=refene1, refene_old=refene1, pka_corr=7.1, # HIP
              charges=[-0.3479, 0.2747, -0.1354, 0.1212, -0.0414, 0.0810,
              0.0810, -0.0012, -0.1513, 0.3866, -0.0170, 0.2681, -0.1718,
              0.3911, -0.1141, 0.2317, 0.7341, -0.5894])
HIP.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=0.6, # HID
              charges=[-0.3479, 0.2747, -0.1354, 0.1212, -0.1110, 0.0402,
              0.0402, -0.0266, -0.3811, 0.3649, 0.2057, 0.1392, -0.5727, 0.0,
              0.1292, 0.1147, 0.7341, -0.5894])
HIP.add_state(protcnt=1, refene=refene3, refene_old=refene3_old, pka_corr=0.0, # HIE
              charges=[-0.3479, 0.2747, -0.1354, 0.1212, -0.1012, 0.0367,
              0.0367, 0.1868, -0.5432, 0.0, 0.1635, 0.1435, -0.2795, 0.3339,
              -0.2207, 0.1862, 0.7341, -0.5894])
HIP.check()

# Lysine
refene1 = _ReferenceEnergy(igb2=-15.2423959, igb5=-14.5392838, igb8=-18.393654)
refene1.solvent_energies(igb2=-15.1417977, igb5=-14.3152107)
refene1.dielc2_energies(igb2=-7.239587, igb5=-6.825997)
refene1.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene1_old = _ReferenceEnergy(igb2=-15.2423959, igb5=-14.5392838, igb8=-18.393654)
refene1_old.solvent_energies(igb2=-15.1417977, igb5=-14.3152107)
refene1_old.dielc2_energies(igb2=-7.239587, igb5=-6.825997)
refene1_old.dielc2.solvent_energies()
refene1_old.set_pKa(10.4, deprotonated=False)
refene2 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene2.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)

LYS = TitratableResidue('LYS', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'HG2', 'HG3', 'CD', 'HD2', 'HD3', 'CE', 'HE2' ,'HE3',
                        'NZ', 'HZ1', 'HZ2', 'HZ3', 'C', 'O'], pka=10.4, typ="ph")
LYS.add_state(protcnt=3, refene=refene1, refene_old=refene1_old, pka_corr=10.4, # protonated
              charges=[-0.3479, 0.2747, -0.2400, 0.1426, -0.0094, 0.0362,
              0.0362, 0.0187, 0.0103, 0.0103, -0.0479, 0.0621, 0.0621, -0.0143,
              0.1135, 0.1135, -0.3854, 0.3400, 0.3400, 0.3400, 0.7341, -0.5894])
LYS.add_state(protcnt=2, refene=refene2, refene_old=refene2, pka_corr=0.0, # deprotonated
              charges=[-0.3479, 0.2747, -0.2400, 0.1426, -0.10961, 0.0340,
              0.0340, 0.06612, 0.01041, 0.01041, -0.03768, 0.01155, 0.01155,
              0.32604, -0.03358, -0.03358, -1.03581, 0.0, 0.38604, 0.38604,
              0.7341, -0.5894])
LYS.check()

# Cysteine
refene1 = _ReferenceEnergy(igb2=77.4666763, igb5=76.2588331, igb8=71.5804519)
refene1.solvent_energies(igb2=77.6041407, igb5=76.2827217)
refene1.dielc2_energies(igb2=38.090523, igb5=37.454637)
refene1.dielc2.solvent_energies(igb2=38.489170)
# Copying the reference energy to be printted on the old CPIN format
refene1_old = _ReferenceEnergy(igb2=77.4666763, igb5=76.2588331, igb8=71.5804519)
refene1_old.solvent_energies(igb2=77.6041407, igb5=76.2827217)
refene1_old.dielc2_energies(igb2=38.090523, igb5=37.454637)
refene1_old.dielc2.solvent_energies(igb2=38.489170)
refene1_old.set_pKa(8.5, deprotonated=False)
refene2 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene2.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)

CYS = TitratableResidue('CYS', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'SG',
                        'HG', 'C', 'O'], pka=8.5, typ="ph")
CYS.add_state(protcnt=1, refene=refene1, refene_old=refene1_old, pka_corr=8.5, # protonated
              charges=[-0.4157, 0.2719, 0.0213, 0.1124, -0.1231, 0.1112, 0.1112,
                       -0.3119, 0.1933, 0.5973, -0.5679])
CYS.add_state(protcnt=0, refene=refene2, refene_old=refene2, pka_corr=0.0, # deprotonated
              charges=[-0.4157, 0.2719, 0.0213, 0.1124, -0.3593, 0.1122, 0.1122,
                       -0.8844, 0.0, 0.5973, -0.5679])
CYS.check()

# Deoxy-adenine
refene1 = _ReferenceEnergy(igb2=-19.8442, igb5=-19.8442)
refene1.solvent_energies()
refene1.dielc2_energies(igb2=-9.106013, igb5=-9.404867)
refene1.dielc2.solvent_energies(igb2=-9.779586)
# Copying the reference energy to be printted on the old CPIN format
refene1_old = _ReferenceEnergy(igb2=-19.8442, igb5=-19.8442)
refene1_old.solvent_energies()
refene1_old.dielc2_energies(igb2=-9.106013, igb5=-9.404867)
refene1_old.dielc2.solvent_energies(igb2=-9.779586)
refene1_old.set_pKa(3.9, deprotonated=True)
refene2 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene2.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)

DAP = TitratableResidue('DAP', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                        "C4'", "H4'", "O4'", "C1'", "H1'", 'N9', 'C8', 'H8',
                        'N7', 'C5', 'C6', 'N6', 'H61', 'H62', 'N1', 'C2', 'H2',
                        'N3', 'C4', "C3'", "H3'", "C2'", "H2'1", "H2'2", "O3'",
                        'H1'], pka=3.9, typ="ph")
DAP.add_state(protcnt=1, refene=refene1, refene_old=refene1_old, pka_corr=0.0, # deprotonated
              charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
              0.0754, 0.1629, 0.1176, -0.3691, 0.0431, 0.1838, -0.0268, 0.1607,
              0.1877, -0.6175, 0.0725, 0.6897, -0.9123, 0.4167, 0.4167, -0.7624,
              0.5716, 0.0598, -0.7417, 0.38, 0.0713, 0.0985, -0.0854, 0.0718,
              0.0718, -0.5232, 0.0])
DAP.add_state(protcnt=2, refene=refene2, refene_old=refene2, pka_corr=3.9, # protonated
              charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
              0.0754, 0.1629, 0.1176, -0.3691, 0.0431, 0.1838, 0.0944, 0.1617,
              0.2281, -0.5674, 0.1358, 0.5711, -0.8251, 0.4456, 0.4456, -0.575,
              0.4251, 0.1437, -0.5611, 0.3421, 0.0713, 0.0985, -0.0854, 0.0718,
              0.0718, -0.5232, 0.4301])
DAP.check()

# Deoxy-cytosine
refene1 = _ReferenceEnergy(igb2=-40.526, igb5=-40.526)
refene1.solvent_energies()
refene1.dielc2_energies(igb2=-19.447553, igb5=-19.842087)
refene1.dielc2.solvent_energies(igb2=-20.121129)
# Copying the reference energy to be printted on the old CPIN format
refene1_old = _ReferenceEnergy(igb2=-40.526, igb5=-40.526)
refene1_old.solvent_energies()
refene1_old.dielc2_energies(igb2=-19.447553, igb5=-19.842087)
refene1_old.dielc2.solvent_energies(igb2=-20.121129)
refene1_old.set_pKa(4.3, deprotonated=True)
refene2 = _ReferenceEnergy(igb2=0, igb5=0)
refene2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene2.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)

DCP = TitratableResidue('DCP', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                        "C4'", "H4'", "O4'", "C1'", "H1'", 'N1', 'C6', 'H6',
                        'C5', 'H5', 'C4', 'N4', 'H41', 'H42', 'N3', 'C2', 'O2',
                        "C3'", "H3'", "C2'", "H2'1", "H2'2", "O3'", 'H3'],
                        pka=4.3, typ="ph")
DCP.add_state(protcnt=1, refene=refene1, refene_old=refene1_old, pka_corr=0.0, # deprotonated
              charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
              0.0754, 0.1629, 0.1176, -0.3691, -0.0116, 0.1963, -0.0339,
              -0.0183, 0.2293, -0.5222, 0.1863, 0.8439, -0.9773, 0.4314, 0.4314,
              -0.7748, 0.7959, -0.6548, 0.0713, 0.0985, -0.0854, 0.0718, 0.0718,
              -0.5232, 0.0])
DCP.add_state(protcnt=2, refene=refene2, refene_old=refene2, pka_corr=4.3, # protonated
              charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
              0.0754, 0.1629, 0.1176, -0.3691, -0.0116, 0.1963, 0.2167, -0.0282,
              0.2713, -0.4162, 0.2179, 0.6653, -0.859, 0.4598, 0.4598, -0.4956,
              0.5371, -0.5028, 0.0713, 0.0985, -0.0854, 0.0718, 0.0718, -0.5232,
              0.4108])
DCP.check()

# Deoxy-guanine
refene1 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-90.0011, igb5=-90.0011)
refene2.solvent_energies()
refene2.dielc2_energies(igb2=-44.031593, igb5=-43.588343)
refene2.dielc2.solvent_energies(igb2=-45.090067)
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=-90.0011, igb5=-90.0011)
refene2_old.solvent_energies()
refene2_old.dielc2_energies(igb2=-44.031593, igb5=-43.588343)
refene2_old.dielc2.solvent_energies(igb2=-45.090067)
refene2_old.set_pKa(9.2, deprotonated=True)

DG = TitratableResidue('DG', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                       "C4'", "H4'", "O4'", "C1'", "H1'", 'N9', 'C8', 'H8',
                       'N7', 'C5', 'C6', 'O6', 'N1', 'H1', 'C2', 'N2', 'H21',
                       'H22', 'N3', 'C4', "C3'", "H3'", "C2'", "H2'1", "H2'2",
                       "O3'"], pka=9.2, typ="ph")
DG.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=9.2, # protonated
             charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
             0.0754, 0.1629, 0.1176, -0.3691, 0.0358, 0.1746, 0.0577, 0.0736,
             0.1997, -0.5725, 0.1991, 0.4918, -0.5699, -0.5053, 0.352, 0.7432,
             -0.923, 0.4235, 0.4235, -0.6636, 0.1814, 0.0713, 0.0985, -0.0854,
             0.0718, 0.0718, -0.5232])
DG.add_state(protcnt=0, refene=refene2, refene_old=refene2_old, pka_corr=0.0, # deprotonated
             charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
             0.0754, 0.1629, 0.1176, -0.3691, 0.0358, 0.1746, -0.0507, 0.0779,
             0.1516, -0.6122, 0.0806, 0.7105, -0.7253, -0.8527, 0.0, 0.9561,
             -0.9903, 0.3837, 0.3837, -0.8545, 0.2528, 0.0713, 0.0985, -0.0854,
             0.0718, 0.0718, -0.5232])
DG.check()

# Deoxy-thymine
refene1 = _ReferenceEnergy(igb2=0, igb5=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-56.7729, igb5=-56.7729)
refene2.solvent_energies(igb2=-28.429391)
refene2.dielc2_energies(igb2=-28.085730, igb5=-27.298290)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=-56.7729, igb5=-56.7729)
refene2_old.solvent_energies(igb2=-28.429391)
refene2_old.dielc2_energies(igb2=-28.085730, igb5=-27.298290)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(9.7, deprotonated=True)

DT = TitratableResidue('DT', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                       "C4'", "H4'", "O4'", "C1'", "H1'", 'N1', 'C6', 'H6',
                       'C5', 'C7', 'H71', 'H72', 'H73', 'C4', 'O4', 'N3', 'H3',
                       'C2', 'O2', "C3'", "H3'", "C2'", "H2'1", "H2'2", "O3'"],
                       pka=9.7, typ="ph")
DT.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=9.7, # protonated
             charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
             0.0754, 0.1629, 0.1176, -0.3691, 0.068, 0.1804, -0.0239, -0.2209,
             0.2607, 0.0025, -0.2269, 0.077, 0.077, 0.077, 0.5194, -0.5563,
             -0.434, 0.342, 0.5677, -0.5881, 0.0713, 0.0985, -0.0854, 0.0718,
             0.0718, -0.5232])
DT.add_state(protcnt=0, refene=refene2, refene_old=refene2_old, pka_corr=0.0, # deprotonated
             charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
             0.0754, 0.1629, 0.1176, -0.3691, 0.068, 0.1804, -0.2861, -0.1874,
             0.2251, -0.1092, -0.2602, 0.0589, 0.0589, 0.0589, 0.8263, -0.7396,
             -0.9169, 0.0, 0.9167, -0.7722, 0.0713, 0.0985, -0.0854, 0.0718,
             0.0718, -0.5232])
DT.check()

# Adenine
refene1 = _ReferenceEnergy(igb2=0, igb5=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=14.851840012, igb5=15.1166001033)
refene2.solvent_energies(igb2=15.026098360790293,igb5=15.143651997474915)
refene2.solvent_energies()
refene2.dielc2_energies(igb2=6.953887, igb5=7.092043)
refene2.dielc2.solvent_energies(igb2=7.544988)
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=14.851840012, igb5=15.1166001033)
refene2_old.solvent_energies(igb2=15.026098360790293,igb5=15.143651997474915)
refene2_old.solvent_energies()
refene2_old.dielc2_energies(igb2=6.953887, igb5=7.092043)
refene2_old.dielc2.solvent_energies(igb2=7.544988)
refene2_old.set_pKa(3.5, deprotonated=False)

AP = TitratableResidue('AP', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                       "C4'", "H4'", "O4'", "C1'", "H1'", 'N9', 'C8', 'H8',
                       'N7', 'C5', 'C6', 'N6', 'H61', 'H62', 'N1', 'C2', 'H2',
                       'N3', 'C4', "C3'", "H3'", "C2'", "H2'1", "O2'", "HO'2",
                       "O3'", 'H1'], pka=3.9, typ="ph")
AP.add_state(protcnt=0, refene=refene1, refene_old=refene1, pka_corr=0.0, # deprotonated
             charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
             0.1065, 0.1174, -0.3548, 0.0394, 0.2007, -0.0251, 0.2006, 0.1553,
             -0.6073, 0.0515, 0.7009, -0.9019, 0.4115, 0.4115, -0.7615, 0.5875,
             0.0473, -0.6997, 0.3053, 0.2022, 0.0615, 0.067, 0.0972, -0.6139,
             0.4186, -0.5246, 0.0])
AP.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=3.5, # protonated
             charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
             0.1065, 0.1174, -0.3548, 0.0394, 0.2007, 0.0961, 0.2011, 0.1965,
             -0.5569, 0.1136, 0.5845, -0.8152, 0.4403, 0.4403, -0.5776, 0.4435,
             0.1307, -0.5201, 0.2681, 0.2022, 0.0615, 0.067, 0.0972, -0.6139,
             0.4186, -0.5246, 0.431])
AP.check()

# Cytosine
refene1 = _ReferenceEnergy(igb2=0, igb5=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=37.501800178, igb5=38.0081251132)
refene2.solvent_energies(igb2=37.378544257354164, igb5=37.90444570773976)
refene2.dielc2_energies(igb2=18.483513, igb5=19.016390)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=37.501800178, igb5=38.0081251132)
refene2_old.solvent_energies(igb2=37.378544257354164, igb5=37.90444570773976)
refene2_old.dielc2_energies(igb2=18.483513, igb5=19.016390)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(4.2, deprotonated=False)

CP = TitratableResidue('CP', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                       "C4'", "H4'", "O4'", "C1'", "H1'", 'N1', 'C6', 'H6',
                       'C5', 'H5', 'C4', 'N4', 'H41', 'H42', 'N3', 'C2', 'O2',
                       "C3'", "H3'", "C2'", "H2'1", "O2'", "HO'2", "O3'", 'H3'],
                       pka=4.3, typ="ph")
CP.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=0.0, # deprotonated
             charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
             0.1065, 0.1174, -0.3548, 0.0066, 0.2029, -0.0484, 0.0053, 0.1958,
             -0.5215, 0.1928, 0.8185, -0.953, 0.4234, 0.4234, -0.7584, 0.7538,
             -0.6252, 0.2022, 0.0615, 0.067, 0.0972, -0.6139, 0.4186, -0.5246,
             0.0])
CP.add_state(protcnt=2, refene=refene2, refene_old=refene2_old, pka_corr=4.2, # protonated
             charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
             0.1065, 0.1174, -0.3548, 0.0066, 0.2029, 0.1954, 0.0028, 0.2366,
             -0.4218, 0.2253, 0.6466, -0.8363, 0.4518, 0.4518, -0.4871, 0.5039,
             -0.4753, 0.2022, 0.0615, 0.067, 0.0972, -0.6139, 0.4186, -0.5246,
             0.4128])
CP.check()

# Guanine
refene1 = _ReferenceEnergy(igb2=0, igb5=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-97.094725165, igb5=-96.0365352027)
refene2.solvent_energies(igb2=-97.31657849010276, igb5=-95.95654436492156)
refene2.dielc2_energies(igb2=-47.410980, igb5=-47.008233)
refene2.dielc2.solvent_energies(igb2=-48.222021)
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=-97.094725165, igb5=-96.0365352027)
refene2_old.solvent_energies(igb2=-97.31657849010276, igb5=-95.95654436492156)
refene2_old.dielc2_energies(igb2=-47.410980, igb5=-47.008233)
refene2_old.dielc2.solvent_energies(igb2=-48.222021)
refene2_old.set_pKa(9.2, deprotonated=True)

G = TitratableResidue('G', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                      "C4'", "H4'", "O4'", "C1'", "H1'", 'N9', 'C8', 'H8', 'N7',
                      'C5', 'C6', 'O6', 'N1', 'H1', 'C2', 'N2', 'H21', 'H22',
                      'N3', 'C4', "C3'", "H3'", "C2'", "H2'1", "O2'", "HO'2",
                      "O3'"], pka=9.2, typ="ph")
G.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=9.2, # protonated
            charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
            0.1065, 0.1174, -0.3548, 0.0191, 0.2006, 0.0492, 0.1374, 0.164,
            -0.5709, 0.1744, 0.477, -0.5597, -0.4787, 0.3424, 0.7657, -0.9672,
            0.4364, 0.4364, -0.6323, 0.1222, 0.2022, 0.0615, 0.067, 0.0972,
            -0.6139, 0.4186, -0.5246])
G.add_state(protcnt=0, refene=refene2, refene_old=refene2_old, pka_corr=0.0, # deprotonated
            charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
            0.1065, 0.1174, -0.3548, 0.0191, 0.2006, -0.0623, 0.1479, 0.1137,
            -0.6127, 0.0488, 0.7137, -0.7191, -0.8557, 0.0, 0.9976, -1.0387,
            0.3969, 0.3969, -0.8299, 0.1992, 0.2022, 0.0615, 0.067, 0.0972,
            -0.6139, 0.4186, -0.5246])
G.check()

# Uracil
refene1 = _ReferenceEnergy(igb2=0, igb5=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-136.326020191, igb5=-134.938275039)
refene2.solvent_energies(igb2=-136.5653533428478, igb5=-135.06973320905044)
refene2.dielc2_energies(igb2=-67.270690, igb5=-66.605330)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=-136.326020191, igb5=-134.938275039)
refene2_old.solvent_energies(igb2=-136.5653533428478, igb5=-135.06973320905044)
refene2_old.dielc2_energies(igb2=-67.270690, igb5=-66.605330)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(9.2, deprotonated=True)

U = TitratableResidue('U', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                      "C4'", "H4'", "O4'", "C1'", "H1'", 'N1', 'C6', 'H6', 'C5',
                      'H5', 'C4', 'O4', 'N3', 'H3', 'C2', 'O2', "C3'", "H3'",
                      "C2'", "H2'1", "O2'", "HO'2", "O3'"], pka=9.3, typ="ph")
U.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=9.2, # protonated
            charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
            0.1065, 0.1174, -0.3548, 0.0674, 0.1824, 0.0418, -0.1126, 0.2188,
            -0.3635, 0.1811, 0.5952, -0.5761, -0.3549, 0.3154, 0.4687, -0.5477,
            0.2022, 0.0615, 0.067, 0.0972, -0.6139, 0.4186, -0.5246])
U.add_state(protcnt=0, refene=refene2, refene_old=refene2_old, pka_corr=0.0, # deprotonated
            charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
            0.1065, 0.1174, -0.3548, 0.0674, 0.1824, -0.2733, 0.0264, 0.1501,
            -0.582, 0.156, 0.9762, -0.7808, -0.9327, 0.0, 0.8698, -0.7435,
            0.2022, 0.0615, 0.067, 0.0972, -0.6139, 0.4186, -0.5246])
U.check()

# HEH: HEME ring + parts of 2 HIS + 2 CYS
refene1 = _ReferenceEnergy(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-15.493731, igb5=-16.349152, igb7=-16.509509, igb8=-22.025653) # Implicit
refene2.solvent_energies(igb2=-15.209270, igb5=-15.840853, igb7=-15.495868) # Explicit
refene2.dielc2_energies()
refene2.dielc2.solvent_energies()

HEH = TitratableResidue('HEH', ['FE', 'NA', 'C1A', 'C2A', 'C3A', 'CMA', 'HMA1', 'HMA2', 'HMA3', 'C4A',
                                'CHB', 'HHB', 'C1B', 'NB', 'C2B', 'CMB', 'HMB1', 'HMB2', 'HMB3', 'C3B',
                                'CAB', 'HAB', 'CBB', 'HBB1', 'HBB2', 'HBB3', 'C4B', 'CHC', 'HHC', 'C1C',
                                'NC', 'C2C', 'CMC', 'HMC1', 'HMC2', 'HMC3', 'C3C', 'CAC', 'HAC', 'CBC',
                                'HBC1', 'HBC2', 'HBC3', 'C4C', 'CHD', 'HHD', 'C1D', 'ND', 'C2D', 'CMD',
                                'HMD1', 'HMD2', 'HMD3', 'C3D', 'C4D', 'CHA', 'HHA', 'CBC1', 'HB2C',
                                'HB3C', 'SGC1', 'CB1', 'HB21', 'HB31', 'CG1', 'ND11', 'HD11', 'CE11',
                                'HE11', 'NE21', 'CD21', 'HD21', 'CBB2', 'HB2B', 'HB3B', 'SGB2', 'CB2',
                                'HB22', 'HB32', 'CG2', 'ND12', 'HD12', 'CE12', 'HE12', 'NE22', 'CD22',
                                'HD22'],
                        eo=-0.203, typ="redox")
HEH.add_state(eleccnt=2, refene=refene1, eo_corr=0.0, # FE3+ (oxidized state)
              charges=[ 0.6660, -0.1530, -0.0956,  0.1274,  0.1624, -0.2600,  0.0743,  0.0743,  0.0743,
                       -0.0766, -0.0586,  0.1300, -0.0206, -0.2560,  0.1394, -0.2240,  0.0663,  0.0663,
                        0.0663,  0.0734, -0.0935,  0.1765, -0.4035,  0.1375,  0.1375,  0.1375,  0.0504,
                       -0.1806,  0.1430,  0.0364, -0.2390,  0.1784, -0.2325,  0.0635,  0.0635,  0.0635,
                        0.0084, -0.0570,  0.2130, -0.4090,  0.1357,  0.1357,  0.1357, -0.0266, -0.0576,
                        0.1360, -0.1156, -0.1130,  0.1604, -0.2575,  0.0752,  0.0752,  0.0752,  0.1444,
                       -0.1126,  0.0104,  0.1090, -0.3349,  0.1297,  0.1297, -0.1760, -0.0803,  0.0269,
                        0.0269,  0.1990, -0.2930,  0.3650,  0.0120,  0.1180, -0.0400, -0.2020,  0.1790,
                       -0.3349,  0.1297,  0.1297, -0.1760, -0.0803,  0.0269,  0.0269,  0.1990, -0.2930,
                        0.3650,  0.0120,  0.1180, -0.0400, -0.2020,  0.1790]
              )
HEH.add_state(eleccnt=3, refene=refene2, eo_corr=-0.203, # FE2+ (reduced state)
              charges=[ 0.4800, -0.1337, -0.1455,  0.1285,  0.1325, -0.2545,  0.0608,  0.0608,  0.0608,
                       -0.0865, -0.0815,  0.1220, -0.0425, -0.2490,  0.1605, -0.1935,  0.0422,  0.0422,
                        0.0422, -0.0025, -0.0390,  0.1720, -0.4255,  0.1348,  0.1348,  0.1348,  0.0405,
                       -0.1725,  0.1320, -0.0525, -0.1980,  0.2125, -0.1985,  0.0405,  0.0405,  0.0405,
                       -0.0355, -0.0195,  0.1915, -0.4340,  0.1320,  0.1320,  0.1320, -0.0275, -0.0805,
                        0.1290, -0.1275, -0.1020,  0.1335, -0.2525,  0.0615,  0.0615,  0.0615,  0.1415,
                       -0.1575,  0.0275,  0.0950, -0.3343,  0.1300,  0.1300, -0.2315, -0.0967,  0.0187,
                        0.0187,  0.2030, -0.3450,  0.3550,  0.0110,  0.1090, -0.0270, -0.2170,  0.1750,
                       -0.3343,  0.1300,  0.1300, -0.2315, -0.0967,  0.0187,  0.0187,  0.2030, -0.3450,
                        0.3550,  0.0110,  0.1090, -0.0270, -0.2170,  0.1750]
              )
HEH.check()

# Propionate
refene1 = _ReferenceEnergy(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=10.356928, igb5=9.943308, igb7=7.020632, igb8=5.259028) # Implicit
refene2.solvent_energies(igb2=16.751825, igb5=15.661934, igb7=12.906876, igb8=15.024553) # Explicit
refene2.dielc2_energies()
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=10.356928, igb5=9.943308, igb7=7.020632, igb8=5.259028) # Implicit
refene2_old.solvent_energies(igb2=16.751825, igb5=15.661934, igb7=12.906876, igb8=15.024553) # Explicit
refene2_old.dielc2_energies()
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(4.85, deprotonated=False)

PRN = TitratableResidue('PRN',
                        ['CA', 'HA1', 'HA2', 'CB', 'HB1', 'HB2', 'CG',
                        'O1', 'O2', 'H11', 'H12', 'H21', 'H22'], pka=4.85, typ="ph")

PRN.add_state(protcnt=0, refene=refene1, refene_old=refene1, pka_corr=0.0, # deprotonated
              charges=[-0.0508, -0.0173,
              -0.0173, 0.0026, -0.0425, -0.0425, 0.8054, -0.8188, -0.8188, 0.0,
              0.0, 0.0, 0.0])

PRN.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.85, # protonated syn-O1
              charges=[-0.0181, 0.0256, 0.0256,
              -0.0284, 0.0430, 0.0430, 0.6801, -0.6511, -0.5838, 0.4641,
              0.0, 0.0, 0.0])

PRN.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.85, # protonated anti-O1
              charges=[-0.0181, 0.0256, 0.0256,
              -0.0284, 0.0430, 0.0430, 0.6801, -0.6511, -0.5838, 0.0,
              0.4641, 0.0, 0.0])

PRN.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.85, # protonated syn-O2
              charges=[-0.0181, 0.0256, 0.0256,
              -0.0284, 0.0430, 0.0430, 0.6801, -0.5838, -0.6511, 0.0,
              0.0, 0.4641, 0.0])

PRN.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.85, # protonated anti-O2
              charges=[-0.0181, 0.0256, 0.0256,
              -0.0284, 0.0430, 0.0430, 0.6801, -0.5838, -0.6511, 0.0,
              0.0, 0.0, 0.4641])

PRN.check()

# Tyrosine, pH and redox active
refene1 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=5.409084) # Implicit
refene2.solvent_energies(igb2=5.210075) # Explicit
refene2.dielc2_energies()
refene2.dielc2.solvent_energies()
refene3 = _ReferenceEnergy(igb2=-52.293859) # Implicit
refene3.solvent_energies(igb2=-52.185005) # Explicit
refene3.dielc2_energies()
refene3.dielc2.solvent_energies()
refene4 = _ReferenceEnergy(igb2=24.166408) # Implicit
refene4.solvent_energies(igb2=24.224913) # Explicit
refene4.dielc2_energies()
refene4.dielc2.solvent_energies()

TYX = TitratableResidue('TYX', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'CD1', 'HD1', 'CE1', 'HE1', 'CZ', 'OH', 'HH', 'CE2',
                        'HE2', 'CD2', 'HD2', 'C', 'O'], typ="phredox")
# Note: pka_corr differences only make sense between states with the same number of electrons
# Note: eo_corr differences only make sense between states with the same number of protons
TYX.add_state(protcnt=1, eleccnt=1, refene=refene1, pka_corr=9.6, eo_corr=1.4, # tyrOH
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.1163, 0.0548,
                       0.0548, -0.0139, -0.1142, 0.1615, -0.3410, 0.1911,
                       0.4198, -0.5278, 0.3621, -0.3410, 0.1911, -0.1142,
                       0.1615, 0.5973, -0.5679])
TYX.add_state(protcnt=1, eleccnt=0, refene=refene2, pka_corr=-2.0, eo_corr=0.0, # tyrOH+
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.2911, 0.1123,
                       0.1123, 0.4479, -0.1923, 0.2177, -0.1736, 0.2109,
                       0.5560, -0.4376, 0.4031, -0.1736, 0.2109, -0.1923,
                       0.2177, 0.5973, -0.5679])
TYX.add_state(protcnt=0, eleccnt=1, refene=refene3, pka_corr=0.0, eo_corr=0.71, # tyrO-
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.0500, 0.0300,
                       0.0300, -0.1786, -0.1545, 0.1418, -0.4793, 0.1318,
                       0.7197, -0.8026, 0.0000, -0.4793, 0.1318, -0.1545,
                       0.1418, 0.5974, -0.5678])
TYX.add_state(protcnt=0, eleccnt=0, refene=refene4, pka_corr=0.0, eo_corr=0.0, # tyrO
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.1941, 0.0931,
                       0.0931, 0.0538, -0.1216, 0.1629, -0.3224, 0.1643,
                       0.6679, -0.4519, 0.0000, -0.3224, 0.1643, -0.1216,
                       0.1629, 0.5973, -0.5679])
TYX.check()



import math
import itertools

from ._edit_descriptors import *
from ._misc import expand_edit_descriptors, has_next_iterator
from . import config

PROC_SIGN_ZERO = config.PROC_SIGN_ZERO
PROC_MIN_FIELD_WIDTH = config.PROC_MIN_FIELD_WIDTH
PROC_DECIMAL_CHAR = config.PROC_DECIMAL_CHAR
G0_NO_BLANKS = config.G0_NO_BLANKS
PROC_NO_LEADING_BLANK = config.PROC_NO_LEADING_BLANK

def output(eds, reversion_eds, values):
    '''
    a function to take a list of valid f77 edit descriptors and respective values
    and output the corresponding string
    '''
    record = ''
    state = {
        'position' : 0,
        'scale' : 0,
        'incl_plus' : False,
        'blanks_as_zeros' : False,
        'halt_if_no_vals' : False,
    }

    # if format is empty with no values specified, then output blank record -
    # see section 13.3

    # check that if there is a reversion, that values can be output
    reversion_contains_output_ed = False
    for ed in reversion_eds:
        if isinstance(ed, OUTPUT_EDS):
            reversion_contains_output_ed = True
            break
    # Get full list of edit descriptors
    eds = expand_edit_descriptors(eds)
    reversion_eds = expand_edit_descriptors(reversion_eds)
    # use iterators
    get_ed = has_next_iterator(eds)
    get_value = has_next_iterator(values)
    tmp_reversion_eds = []
    # continue until out of edit descriptors or values
    while True:
        # no more edit descriptors, no more values, stop output, 
        # todo: this will cut short a reversion edit descriptor section - is this right?
        if not get_ed.has_next() and not get_value.has_next():
            break
        # take a edit descriptor off the queue if there is any
        if get_ed.has_next():
            ed = next(get_ed)
        else:
            if reversion_contains_output_ed == True:
                # take from reversion edit descriptors if there is a value
                # requiring output still
                while True:
                    if len(tmp_reversion_eds):
                        ed = tmp_reversion_eds.pop()
                        # these edit descriptors are ignored in reversion state
                        if not isinstance(ed, NON_REVERSION_EDS):
                            break
                    else:
                        # Regardless of where cursor is, is moved to the
                        # next record
                        record = record + config.RECORD_SEPARATOR
                        state['position'] = len(record)
                        tmp_reversion_eds = reversion_eds[::-1]
            else:
                # ignore the revsion edit descriptors as cannot output the
                # final value
                break
        # check if edit descriptor requires a value
        if isinstance(ed, OUTPUT_EDS):
            if get_value.has_next():
                val = next(get_value)
            else:
                # is a edit descriptor that requires a value but no value
                # todo: does it stop gracefully or raise error?
                break   
            if isinstance(ed, I):
                sub_string = _compose_i_string(ed.width, ed.min_digits, state, val)
            elif isinstance(ed, B):
                w = ed.width
                m = ed.min_digits
                sub_string = _compose_boz_string(w, m, state, val, 'B')
            if isinstance(ed, O):
                w = ed.width
                m = ed.min_digits
                sub_string = _compose_boz_string(w, m, state, val, 'O')
            if isinstance(ed, Z):
                w = ed.width
                m = ed.min_digits
                sub_string = _compose_boz_string(w, m, state, val, 'Z')
            elif isinstance(ed, F):
                w = ed.width
                e = None
                d = ed.decimal_places
                sub_string = _compose_float_string(w, e, d, state, val, 'F')
            elif isinstance(ed, E):
                w = ed.width
                e = ed.exponent
                d = ed.decimal_places
                sub_string = _compose_float_string(w, e, d, state, val, 'E')
            elif isinstance(ed, D):
                w = ed.width
                e = None
                d = ed.decimal_places
                sub_string = _compose_float_string(w, e, d, state, val, 'D')
            elif isinstance(ed, G):
                w = ed.width
                e = ed.exponent
                d = ed.decimal_places
                sub_string = _compose_float_string(w, e, d, state, val, 'G')
            elif isinstance(ed, EN):
                w = ed.width
                e = ed.exponent
                d = ed.decimal_places
                sub_string = _compose_float_string(w, e, d, state, val, 'EN')
            elif isinstance(ed, ES):
                w = ed.width
                e = ed.exponent
                d = ed.decimal_places
                sub_string = _compose_float_string(w, e, d, state, val, 'ES')
            elif isinstance(ed, L):
                sub_string = _compose_l_string(ed.width, state, val)
            elif isinstance(ed, A):
                sub_string = _compose_a_string(ed.width, state, val)
            state['position'], record = _write_string(record, sub_string, state['position'])
        else:
            # token does not require a value
            if isinstance(ed, (S, SS)):
                state['incl_plus'] = False
            if isinstance(ed, SP):
                state['incl_plus'] = True
            elif isinstance(ed, P):
                state['scale'] = ed.scale
            elif isinstance(ed, BN):
                # This is moot since for output, this does not do anything
                state['blanks_as_zeros'] = False
            elif isinstance(ed, BZ):
                state['blanks_as_zeros'] = True
            elif isinstance(ed, Colon):
                state['halt_if_no_vals'] = True
            elif isinstance(ed, Slash):
                state['position'], record = _write_string(record, config.RECORD_SEPARATOR, state['position'])
            elif isinstance(ed, (X, TR)):
                state['position'] = state['position'] + ed.num_chars
            elif isinstance(ed, TL):
                state['position'] = state['position'] - ed.num_chars
            elif isinstance(ed, T):
                state['position'] = ed.num_chars - 1
            elif isinstance(ed, QuotedString):
                sub_string = ed.char_string
                state['position'], record = _write_string(record, sub_string, state['position'])
    # output the final record
    return record

def _compose_nan_string(w, ftype):
    if ftype in ['B', 'O', 'Z']:
        return ''
    else:
        # Allow at least 'NaN' to be printed
        if w == 0:
            w = 4 # n.b. this is what is set in Gfortran 4.4.0
        if w < 3:
            return '*' * w
    return 'NaN'.rjust(w)

def _compose_inf_string(w, ftype, sign_bit):
    if ftype in ['B', 'O', 'Z']:
        return ''
    else:
        sign = '+'
        # Allow at least 'Inf' to be printed
        if w == 0:
            w = 4
        if w < 3:
            return '*' * w
        # Change sign if negative
        if sign_bit:
            sign = '-'
            # Require sign if negative, if no space then overflow
            if w == 3:
                return '*' * w
        # Output long version if long enough
        if w > 8:
            return (sign + 'Infinity').rjust(w)
        # Output shortened version with sign if long enough
        elif w > 3:
            return (sign + 'Inf').rjust(w)
        # Should only output short version with no sign if positive
        else:
            return 'Inf'
                

def _compose_float_string(w, e, d, state, val, ftype):
    '''
    Adapted from code in glibfortran which is written in C so is somwhat
    'bit-pushy' in nature. Writes the value to an initial string (buffer)
    and then pulls the subsequent strings from that
    '''
    if (d < 0) or (d is None):
        raise InvalidFormat('Unspecified precision')
    # Make sure they are ints
    d = int(round(d))
    if e is not None:
        e = int(round(e))
    if w is not None:
        w = int(round(w))
    # ==== write_float ==== (function)
    edigits = 4 # Largest number of exponent digits expected
    if (ftype in ['F', 'EN', 'G']) or \
        ((ftype in ['D', 'E']) and (state['scale'] != 0)):
        # Convert with full possible precision
        ndigits = PROC_MIN_FIELD_WIDTH - 4 - edigits
    else:
        # Otherwise convert knowing what the required precision is (i.e. knowing d)
        if ftype == 'ES':
            ndigits = d + 1
        else:
            ndigits = d
        if ndigits > (PROC_MIN_FIELD_WIDTH - 4 - edigits):
            ndigits = PROC_MIN_FIELD_WIDTH - 4 - edigits
    # ==== WRITE_FLOAT ==== (macro)
    # Determine sign of value
    if val == 0.0:
        sign_bit = '-' in str(val)
    else:
        sign_bit = val < 0
    # handle the nan and inf cases
    if type(val) is float and val != val :
        return _compose_nan_string(w, ed)
    Infinity = 1e1000000
    if val in (-Infinity, Infinity):
        return _compose_inf_string(w, ed, sign_bit)
    tmp = abs(val)
    # Round the input if the input is less than 1
    if (ftype == 'F') and (d == state['scale']) and (d == 0):
        if tmp < 1.0:
            tmp = round(tmp)
    zero_flag = (tmp == 0)
    # === DTOA === (macro)
    # write the tmp value to the string buffer
    # sprintf seems to allow negative number of decimal places, need to correct for this
    if ndigits <= 0:
        fmt = '%+-#' + str(PROC_MIN_FIELD_WIDTH) + 'e'
    else:
        fmt = '%+-#' + str(PROC_MIN_FIELD_WIDTH) + '.' + str(ndigits - 1) + 'e'
    buff = fmt % tmp
    # === WRITE_FLOAT === (macro)
    if ftype != 'G':
        return  _output_float(w, d, e, state, ftype, buff, sign_bit, zero_flag, ndigits, edigits)
    else:
        # Perform different actions for G edit descriptors depending on value
        #
        # Generate corresponding I/O format for FMT_G and output.
        # The rules to translate FMT_G to FMT_E or FMT_F from DEC fortran
        # LRM (table 11-2, Chapter 11, "I/O Formatting", P11-25) is:
        # 
        # Data Magnitude                              Equivalent Conversion
        # 0< m < 0.1-0.5*10**(-d-1)                   Ew.d[Ee]
        # m = 0                                       F(w-n).(d-1), n' '
        # 0.1-0.5*10**(-d-1)<= m < 1-0.5*10**(-d)     F(w-n).d, n' '
        # 1-0.5*10**(-d)<= m < 10-0.5*10**(-d+1)      F(w-n).(d-1), n' '
        # 10-0.5*10**(-d+1)<= m < 100-0.5*10**(-d+2)  F(w-n).(d-2), n' '
        # ................                           ..........
        # 10**(d-1)-0.5*10**(-1)<= m <10**d-0.5       F(w-n).0,n(' ')
        # m >= 10**d-0.5                              Ew.d[Ee]
        # 
        # notes: for Gw.d ,  n' ' means 4 blanks
        #        for Gw.dEe, n' ' means e+2 blanks
        nb = 0
        save_scale_factor = state['scale']
        exp_d = 10 ** d
        if (0.0 <= tmp < (0.1 - 0.05 / exp_d)) or \
            (tmp >= (exp_d - 0.5)):
            ftype = 'E'
        else:
            mag = int(abs(round(math.log10(tmp))))
            low = lambda mag, d : 10 ** (mag - 1) - 5 * 10 ** (-d - 1 + mag) 
            high = lambda mag, d : 10 ** mag - 0.5 * 10 ** (-d + mag)
            while tmp < low(mag, d):
                mag = mag - 1
            while tmp >= high(mag, d):
                mag = mag + 1
            assert(low(mag, d) <= tmp < high(mag, d))
            if e < 0:
                nb = 4
            else:
                nb = e + 2
            ftype = 'F'
            w = w - nb
            if tmp == 0.0:
                d = d - 1
            else:
                d = d - mag
                # d = -(mid - d - 1)
            state['scale'] = 0
        out = _output_float(w, d, e, state, ftype, buff, sign_bit, zero_flag, ndigits, edigits)
        state['scale'] = save_scale_factor
        # TODO: this may not be right ...
        if nb > 0:
            if '*' in out:
                out = out + ('*' * nb)
            else:
                out = out + (' ' * nb)
        if len(out) > (w + nb):
            out = '*' * (w + nb)
        return out
        

def _output_float(w, d, e, state, ft, buff, sign_bit, zero_flag, ndigits, edigits):

    # nbefore - number of digits before the decimal point
    # nzero - number of zeros after the decimal point
    # nafter - number of digits after the decimal point
    # nzero_real - number of zeros after the decimal point regardles of the precision

    # Some hacks to change None to -1 (C convention)
    if w is None:
        w = -1
    if e is None:
        e = -1
    nzero_real = -1
    sign = _calculate_sign(state, sign_bit)
    # Some debug
    if d != 0:
        assert(buff[2] in ['.', ','])
        assert(buff[ndigits + 2] == 'e')
    # Read in the exponent
    ex = int(buff[ndigits + 3:]) + 1
    # Handle zero case
    if zero_flag:
        ex = 0
        if PROC_SIGN_ZERO:
            sign = _calculate_sign(state, sign_bit)
        else:
            sign = _calculate_sign(state, False)
        # Handle special case
        if w == 0:
            w = d + 2
        # This case does not include a decimal point
        if (w == 1) and (ft == 'F'):
            if state['incl_plus']:
                return '*' # This is ifort behaviour
            else:
                return '.' # CHANGED: Was '0'
    # Get rid of the decimal and the initial sign i.e. normalise the digits
    digits = buff[1] + buff[3:]
    # Find out where to place the decimal point
    if ft == 'F':
        nbefore = ex + state['scale']
        if nbefore < 0:
            nzero = -nbefore
            nzero_real = nzero
            if nzero > d:
                nzero = d
            nafter = d - nzero
            nbefore = 0
        else:
            nzero = 0
            nafter = d
        expchar = None
    elif ft in ['E', 'D']:
        i = state['scale']
        if (d <= 0) and (i == 0):
            raise InvalidFormat("Precision not greater than zero in format specifier 'E' or 'D'")
        if (i <= -d) or (i >= (d + 2)):
            raise InvalidFormat("Scale factor out of range in format specifier 'E' or 'D'")
        if not zero_flag:
            ex = ex - i
        if i < 0:
            nbefore = 0
            nzero = -i;
            nafter = d + i
        elif i > 0:
            nbefore = i
            nzero = 0
            nafter = (d - i) + 1
        else:
            nbefore = 0
            nzero = 0
            nafter = d
        expchar = ft
    elif ft == 'EN':
        # Exponent must be a multiple of 3 with 1-3 digits before the d.p
        if not zero_flag:
            ex = ex - 1
        if ex >= 0:
            nbefore = ex % 3
        else:
            nbefore = (-ex) % 3
            if nbefore != 0:
                nbefore = 3 - nbefore
        ex = ex - nbefore
        nbefore = nbefore + 1
        nzero = 0
        nafter = d
        expchar = 'E'
    elif ft == 'ES':
        if not zero_flag:
            ex = ex - 1
        nbefore = 1
        nzero = 0
        nafter = d
        expchar = 'E'
    # Round the value
    if (nbefore + nafter) == 0:
        ndigits = 0
        if (nzero_real == d) and (int(digits[0]) >= 5): # n.b. character comparison not very pythonic!
            # We rounded to zero but shouldn't have
            nzero = nzero - 1
            nafter = 1
            digits = '1' + digits[1:]
            ndigits = 1
    elif (nbefore + nafter) < ndigits:
        ndigits = nbefore + nafter
        i = ndigits
        if int(digits[i]) >= 5:
            # Propagate the carry
            i = i - 1
            while i >= 0:
                digit = int(digits[i])
                if digit != 9:
                    digits = _swapchar(digits, i, str(digit + 1))
                    break
                else:
                    digits = _swapchar(digits, i, '0')
                i = i - 1
            # Did the carry overflow?
            if i < 0:
                digits = '1' + digits
                if ft == 'F':
                    if nzero > 0:
                        nzero = nzero - 1
                        nafter = nafter + 1
                    else:
                        nbefore = nbefore + 1
                elif ft == 'EN':
                    nbefore = nbefore + 1
                    if nbefore == 4:
                        nbefore = 1
                        ex = ex + 3
                else:
                    ex = ex + 1
    # Calculate the format of the exponent field
    if expchar is not None:
        # i = abs(ex)
        # while i >= 10:
        #     edigits = edigits + 1
        #     i = i / 10.0
        if e < 0:
            # Width not specified, must be no more than 3 digits
            if (ex > 999) or (ex < -999):
                edigits = -1
            else:
                edigits = 4
                if (ex > 99) or (ex < -99):
                    expchar = ' '
        else:
            assert(isinstance(ex, int))
            edigits = len(str(abs(ex)))
            # Exponenet width specified, check it is wide enough
            if edigits > e:
                edigits = -1
            else:
                edigits = e + 2
    else:
        edigits = 0
    # Zero values always output as positive, even if the value was egative before rounding
    i = 0
    while i < ndigits:
        if digits[i] != '0':
            break
        i = i + 1
    if i == ndigits:
        # The output is zero so set sign accordingly
        if PROC_SIGN_ZERO:
            sign = _calculate_sign(state, sign_bit)
        else:
            sign = _calculate_sign(state, False)
    # Pick a field size if none was specified
    if w <= 0:
        w = nbefore + nzero + nafter + 1 + len(sign)
    # Work out how much padding is needed
    nblanks = w - (nbefore + nzero + nafter + edigits + 1)
    if sign != '':
        nblanks = nblanks - 1
    # TODO: Find out what this is
    if G0_NO_BLANKS: # dtp->u.p.g0_no_blanks
        w = w - nblanks
        nblanks = 0
    # Check value fits in specified width
    if (nblanks < 0) or (edigits == -1): 
        return '*' * w
    # See if we have space for a zero before the decimal point
    if (nbefore == 0) and (nblanks > 0):
        leadzero = True
        nblanks = nblanks - 1
    else:
        leadzero = False
    out = ''
    # Pad to full field width
    if (nblanks > 0) and not PROC_NO_LEADING_BLANK: # dtp->u.p.no_leading_blank
        out = out + ' ' * nblanks
    # Attach the sign
    out = out + sign
    # Add the lead zero if necessary
    if leadzero:
        out = out + '0'
    # Output portion before the decimal point padded with zeros
    if nbefore > 0:
        if nbefore > ndigits:
            out = out + digits[:ndigits] + (' ' * (nbefore - ndigits))
            digits = digits[ndigits:]
            ndigits = 0
        else:
            i = nbefore
            out = out + digits[:i]
            digits = digits[i:]
            ndigits = ndigits - i
    # Output the decimal point
    out = out + PROC_DECIMAL_CHAR
    # Output the leading zeros after the decimal point
    if nzero > 0:
        out = out + ('0' * nzero)
    # Output the digits after the decimal point, padded with zeros
    if nafter > 0:
        if nafter > ndigits:
            i = ndigits
        else:
            i = nafter
        zeros = '0' * (nafter - i)
        out = out + digits[:i] + zeros
        digits = digits[nafter:]
        ndigits = ndigits - nafter
    # Output the exponent
    if expchar is not None:
        if expchar != ' ':
            out = out + expchar
            edigits = edigits - 1
        fmt = '%+0' + str(edigits) + 'd'
        tmp_buff = fmt % ex
        # if not state['blanks_as_zeros']:
        if PROC_NO_LEADING_BLANK:
            tmp_buf = tmp_buff + (nblanks * ' ')
        out = out + tmp_buff
    return out


def _calculate_sign(state, negative_flag):
    s = ''
    if negative_flag:
        s = '-'
    elif state['incl_plus']:
        s = '+'
    else:
        s = ''
    return s

def _swapchar(s, ind, newch):
    '''
    Helper function to make chars in a string mutableish
    '''
    if 0 < ind >= len(s):
        raise IndexError('index out of range')
    return s[:ind] + newch + s[ind+1:]
    

def _compose_a_string(w, state, val):
    # f77 spec 13.5.11 covers a editing
    val = str(val)
    if w is None:
        output = val
    elif w >= len(val):
        output = val.rjust(w)
    else:
        output = val[:w]
    return output


def _compose_l_string(w, state, val):
    # f77 spec 13.5.10 covers l editing
    try:
        val = bool(val)
    except ValueError:
        raise ValueError("cannot convert '%s' to a boolean" % str(val))
    # single t or f
    if val == True:
        sub_string = 'T'
    else:
        sub_string = 'F'
    # now pad to the specified width
    sub_string = sub_string.rjust(w)
    return sub_string


def _compose_i_string(w, m, state, val):
    # f77 spec 13.5.9.1 covers integer editing 
    null_field = False
    # be pythonic in what values to accept, if it looks like an integer, then
    # so be it
    try:
        val = int(val)
    except ValueError:
        raise ValueError("cannot convert '%s' to a integer" % str(val))
    # get the basic string without sign etc.
    int_string = '%d' % int(round(math.fabs(val)))
    # pad if necessary
    if m is not None:
        int_string = left_pad(int_string, m, '0')
        # weird case where if zero width specified and is zero, can have zero space
        if (val == 0) and (m == 0):
            int_string = ''
    # prepend the sign
    if int_string != '':
        int_string = _get_sign(val, state['incl_plus']) + int_string
    # fill the field with blanks if the number takes more room than the width
    # see f77 spec 13.5.9 remark 5.
    if len(int_string) > w:
        int_string = '*' * w
    else:
        int_string = int_string.rjust(w)
    return int_string

def _get_sign(val, incl_plus):
    if val >= 0:
        if incl_plus == True:
            return '+'
        else:
            return ''
    else:
        return '-'

def _compose_boz_string(w, m, state, val, ftype):
    try:
        val = int(val)
    except ValueError:
        raise ValueError("Cannot convert %s to an integer" % str(val))
    # Special case for zero
    if val == 0:
        if m is None:
            return '0'.rjust(w)
        elif w == m == 0:
            return ' '
        elif m == 0:
            return w * ' '
        else:
            s = left_pad('0', m, '0').rjust(w)
            if len(s) > w:
                return w * '*'
    # Normal cases
    s = ''
    if ftype == 'B':
        # Binary case
        if val < 0:
            return '*' * w
        while val > 0:
            s = str(val % 2) + s
            val = val >> 1
    elif ftype == 'O':
        # Octal case
        if val < 0:
            return '*' * w
        s = '%o' % val
    elif ftype == 'Z':
        # Hex case
        # Fortran uses Two's complement to represent negative numbers, this
        # restricts hex values to sys.maxint, if greater than this overflow,
        # however this is not perfoect as sys.maxint is Pythons build maximum
        # rather than the systems bit size which is difficult to find reliably
        # across platforms
        if abs(val) > config.PROC_MAXINT:
            return '*' * w
        if val < 0:
            s = '%X' % ((config.PROC_MAXINT * 2) + 2 + val)
        else:
            s = '%X' % val
    if m is None:
        s = s.rjust(w)
    else:
        s = left_pad(s, m, '0').rjust(w)
    if len(s) > w:
        return w * '*'
    else:
        return s

def _write_string(record, sub_string, pos):
    '''Function that actually writes the generated strings to a 'stream'''''
    new_pos = pos + len(sub_string)
    # pad if required with blanks - i.e. input after a tr edit descriptor - see
    # f77 format sec. 13.5.3
    if pos > len(record):
        record = record.ljust(pos)
        out =  record + sub_string
    elif pos == len(record):
        out = record + sub_string
    elif pos < len(record):
        out = record[:pos] + sub_string + record[new_pos:]
    return (new_pos, out)


def left_pad(sub_string, width, pad_char):
    # Python 2.3 does not have the character argument to rjust
    padding = pad_char * (width - len(sub_string))
    return padding + sub_string




"""
This is a generalization of the readparm.AmberParm class to handle similar
Amber-style files with %FLAG/%FORMAT tags
"""
from __future__ import division, print_function

from parmed.constants import (NATOM, NTYPES, NBONH, NTHETH, NPHIH,
            NEXT, NRES, NBONA, NTHETA, NPHIA, NUMBND, NUMANG, NPTRA, NATYP,
            NPHB, IFBOX, IFCAP, AMBER_ELECTROSTATIC, CHARMM_ELECTROSTATIC)
from parmed.exceptions import AmberError
from parmed.formats.registry import FileFormatType
from parmed.utils.io import genopen
from parmed.utils.six import string_types, add_metaclass
from parmed.utils.six.moves import range
from contextlib import closing
from copy import copy
import datetime
from parmed.utils.fortranformat import FortranRecordReader, FortranRecordWriter
from math import ceil
import re

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class FortranFormat(object):
    """
    Processes Fortran format strings according to the Fortran specification for
    such formats. This object handles reading and writing data with any valid
    Fortran format. It does this by using the `fortranformat` project
    [https://bitbucket.org/brendanarnold/py-fortranformat].

    However, while `fortranformat` is very general and adheres well to the
    standard, it is very slow. As a result, simple, common format strings have
    been optimized and processes reads and writes between 3 and 5 times faster.
    The format strings (case-insensitive) of the following form (where # can be
    replaced by any number) are optimized:
        - #E#.#
        - #D#.#
        - #F#.#
        - #(F#.#)
        - #a#
        - #I#

    Parameters
    ----------
    format_string : str
        The Fortran Format string to process
    strip_strings : bool=True
        If True, strings are stripped before being processed by stripping
        (only) trailing whitespace
    """

    strre = re.compile(r'(\d+)?a(\d+)$', re.I)
    intre = re.compile(r'(\d+)?i(\d+)$', re.I)
    floatre = re.compile(r'(\d+)?[edf](\d+)\.(\d+)$', re.I)
    floatre2 = re.compile(r'(\d+)?\([edf](\d+)\.(\d+)\)$', re.I)

    #===================================================

    def __init__(self, format_string, strip_strings=True):
        """
        Sets the format string and determines how we will read and write
        strings using this format
        """
        self.format = format_string
        self.strip_strings = strip_strings # for ease of copying

        # Define a function that processes all arguments prior to adding them to
        # the returned list. By default, do nothing, but this allows us to
        # optionally strip whitespace from strings.
        self.process_method = lambda x: x

        if FortranFormat.strre.match(format_string):
            rematch = FortranFormat.strre.match(format_string)
            # replace our write() method with write_string to force left-justify
            self.type, self.write = str, self._write_string
            nitems, itemlen = rematch.groups()
            if nitems is None:
                self.nitems = 1
            else:
                self.nitems = int(nitems)
            self.itemlen = int(itemlen)
            self.fmt = '%s'
            # See if we want to strip the strings
            if strip_strings: self.process_method = lambda x: x.strip()

        elif FortranFormat.intre.match(format_string):
            self.type = int
            rematch = FortranFormat.intre.match(format_string)
            nitems, itemlen = rematch.groups()
            if nitems is None:
                self.nitems = 1
            else:
                self.nitems = int(nitems)
            self.itemlen = int(itemlen)
            self.fmt = '%%%dd' % self.itemlen

        elif FortranFormat.floatre.match(format_string):
            self.type = float
            rematch = FortranFormat.floatre.match(format_string)
            nitems, itemlen, num_decimals = rematch.groups()
            if nitems is None:
                self.nitems = 1
            else:
                self.nitems = int(nitems)
            self.itemlen = int(itemlen)
            self.num_decimals = int(num_decimals)
            if 'F' in format_string.upper():
                self.fmt = '%%%s.%sF' % (self.itemlen, self.num_decimals)
            else:
                self.fmt = '%%%s.%sE' % (self.itemlen, self.num_decimals)

        elif FortranFormat.floatre2.match(format_string):
            self.type = float
            rematch = FortranFormat.floatre2.match(format_string)
            nitems, itemlen, num_decimals = rematch.groups()
            if nitems is None:
                self.nitems = 1
            else:
                self.nitems = int(nitems)
            self.itemlen = int(itemlen)
            self.num_decimals = int(num_decimals)
            if 'F' in format_string.upper():
                self.fmt = '%%%s.%sF' % (self.itemlen, self.num_decimals)
            else:
                self.fmt = '%%%s.%sE' % (self.itemlen, self.num_decimals)

        else:
            # We tried... now just use the fortranformat package
            self._reader = FortranRecordReader(format_string)
            self._writer = FortranRecordWriter(format_string)
            self.write = self._write_ffwriter
            self.read = self._read_ffreader

    #===================================================

    def __copy__(self):
        return type(self)(self.format, self.strip_strings)

    #===================================================

    def __str__(self):
        return self.format

    def __repr__(self):
        return "<%s: %s>" % (type(self).__name__, self.format)

    #===================================================

    def write(self, items, dest):
        """
        Writes an iterable of data (or a single item) to the passed file-like
        object

        Parameters
        ----------
        items : iterable or single float/str/int
            These are the objects to write in this format. The types of each
            item should match the type specified in this Format for that
            argument
        dest : file or file-like
            This is the file to write the data to. It must have a `write` method
            or an AttributeError will be raised

        Notes
        -----
        This method may be replaced with _write_string (for #a#-style formats)
        or _write_ffwriter in the class initializer if no optimization is
        provided for this format, but the call signatures and behavior are the
        same for each of those functions.
        """
        if hasattr(items, '__iter__') and not isinstance(items, string_types):
            mod = self.nitems - 1
            for i, item in enumerate(items):
                dest.write(self.fmt % item)
                if i % self.nitems == mod:
                    dest.write('\n')
            if i % self.nitems != mod:
                dest.write('\n')
        else:
            dest.write(self.fmt % items)
            dest.write('\n')

    #===================================================

    def _write_string(self, items, dest):
        """ Writes a list/tuple of strings """
        if hasattr(items, '__iter__') and not isinstance(items, string_types):
            mod = self.nitems - 1
            for i, item in enumerate(items):
                dest.write((self.fmt % item).ljust(self.itemlen))
                if i % self.nitems == mod:
                    dest.write('\n')
            if i % self.nitems != mod:
                dest.write('\n')
        else:
            dest.write((self.fmt % items).ljust(self.itemlen))
            dest.write('\n')

    #===================================================

    def _read_nostrip(self, line):
        """
        Reads the line and returns converted data. Special-cased for flags that
        may contain 'blank' data. ugh.
        """
        line = line.rstrip('\n')
        nitems = int(ceil(len(line) / self.itemlen))
        ret = [0 for i in range(nitems)]
        start, end = 0, self.itemlen
        for i in range(nitems):
            ret[i] = self.process_method(self.type(line[start:end]))
            start = end
            end += self.itemlen
        return ret

    #===================================================

    def read(self, line):
        """ Reads the line and returns the converted data """
        line = line.rstrip()
        nitems = int(ceil(len(line) / self.itemlen))
        ret = [0 for i in range(nitems)]
        start, end = 0, self.itemlen
        for i in range(nitems):
            ret[i] = self.process_method(self.type(line[start:end]))
            start = end
            end += self.itemlen
        return ret

    #===================================================

    def _read_ffreader(self, line):
        """ Reads the line and returns the converted data """
        return self._reader.read(line.rstrip())

    #===================================================

    def _write_ffwriter(self, items, dest):
        dest.write('%s\n' % self._writer.write(items))

    #===================================================

    def __eq__(self, other):
        return (self.format == other.format and
                self.strip_strings == other.strip_strings)

    #===================================================

    def __hash__(self):
        return hash((self.format, self.strip_strings))

    #===================================================

    def __getstate__(self):
        return dict(format=self.format, strip_strings=self.strip_strings)

    def __setstate__(self, d):
        self.__init__(d['format'], d['strip_strings'])

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@add_metaclass(FileFormatType)
class AmberFormat(object):
    """
    A class that can parse and print files stored in the Amber topology or MDL
    format. In particular, these files have the general form:

    ```
    %VERSION VERSION_STAMP = V00001.000  DATE = XX/XX/XX  XX:XX:XX
    %FLAG <FLAG_NAME>
    %COMMENT <comments>
    %FORMAT(<Fortran_Format>)
    ... data corresponding to that Fortran Format
    %FLAG <FLAG_NAME2>
    %COMMENT <comments>
    %FORMAT(<Fortran_Format>)
    ... data corresponding to that Fortran Format
    ```

    where the `%COMMENT` sections are entirely optional

    Parameters
    ----------
    fname : str=None
        If provided, this file is parsed and the data structures will be loaded
        from the data in this file

    Attributes
    ----------
    parm_data : dict {str : list}
        A dictionary that maps FLAG names to all of the data contained in that
        section of the Amber file.
    formats : dict {str : FortranFormat}
        A dictionary that maps FLAG names to the FortranFormat instance in which
        the data is stored in that section
    parm_comments : dict {str : list}
        A dictionary that maps FLAG names to the list of COMMENT lines that were
        stored in the original file
    flag_list : list
        An ordered list of all FLAG names. This must be kept synchronized with
        `parm_data`, `formats`, and `parm_comments` such that every item in
        `flag_list` is a key to those 3 dicts and no other keys exist
    charge_flag : str='CHARGE'
        The name of the name of the FLAG that describes partial atomic charge
        data. If this flag is found, then its data are multiplied by the
        ELECTROSTATIC_CONSTANT to convert back to fractions of electrons
    version : str
        The VERSION string from the Amber file
    name : str
        The file name of the originally parsed file (set to the fname parameter)
    """
    #===================================================

    @staticmethod
    def id_format(filename):
        """
        Identifies the file type as either Amber-format file (like prmtop) or an
        old-style topology file.

        Parameters
        ----------
        filename : str
            Name of the file to check format for

        Returns
        -------
        is_fmt : bool
            True if it is an Amber-style format, False otherwise
        """
        if isinstance(filename, string_types):
            with closing(genopen(filename, 'r')) as f:
                lines = [f.readline() for i in range(5)]
        elif (hasattr(filename, 'readline') and hasattr(filename, 'seek')
              and hasattr(filename, 'tell')):
            cur = filename.tell()
            lines = [filename.readline() for i in range(5)]
            filename.seek(cur)

        if lines[0].startswith('%VERSION'):
            return True
        # Try old-style format
        try:
            return AmberFormat().rdparm_old(lines, check=True)
        except ValueError:
            return False

    #===================================================

    @staticmethod
    def parse(filename, *args, **kwargs):
        """
        Meant for use with the automatic file loader, this will automatically
        return a subclass of AmberFormat corresponding to what the information
        in the prmtop file contains (i.e., either an AmberParm, ChamberParm,
        AmoebaParm, or AmberFormat)
        """
        from parmed.amber import LoadParm, BeemanRestart
        try:
            return LoadParm(filename, *args, **kwargs)
        except (IndexError, KeyError):
            parm = AmberFormat(filename, *args, **kwargs)
            if 'ATOMIC_COORDS_LIST' in parm.parm_data:
                return BeemanRestart.from_rawdata(parm)
            return parm

    #===================================================

    def __init__(self, fname=None):
        """ Constructor.  Read a file if given """
        self._ncopies = 0
        self.parm_data = {}
        self.formats = {}
        self.parm_comments = {}
        self.flag_list = []
        self.version = None
        self.charge_flag = 'CHARGE'
        self.name = fname

        if fname is not None:
            self.rdparm(fname)

    #===================================================

    def __copy__(self):
        """ Copy all of the data """
        self._ncopies += 1
        other = type(self)()
        other.flag_list = self.flag_list[:]
        other.version = self.version
        if self.name is not None:
            other.name = self.name + '_copy%d' % self._ncopies
        else:
            other.name = None
        other.charge_flag = self.charge_flag
        other.parm_data = {}
        other.parm_comments = {}
        other.formats = {}
        for flag in other.flag_list:
            other.parm_data[flag] = self.parm_data[flag][:]
            other.parm_comments[flag] = self.parm_comments[flag][:]
            other.formats[flag] = copy(self.formats[flag])
        return other

    #===================================================

    def view_as(self, cls):
        """
        Returns a view of the current object as another object.

        Parameters
        ----------
        cls : type
            Class definition of an AmberParm subclass for the current object to
            be converted into

        Returns
        -------
        instance of cls initialized from data in this object. This is NOT a deep
        copy, so modifying the original object may modify this. The copy
        function will create a deep copy of any AmberFormat-derived object
        """
        # If these are the same classes, just return the original instance,
        # since there's nothing to do. Classes are singletons, so use "is"
        if type(self) is cls:
            return self
        return cls.from_rawdata(self)

    #===================================================

    def rdparm(self, fname, slow=False):
        """ Parses the Amber format file """
        self.name = fname
        self.version = None # reset all top info each time rdparm is called
        self.formats = {}
        self.parm_data = {}
        self.parm_comments = {}
        self.flag_list = []

        # See if we have the optimized parser available
        try:
            from parmed.amber import _rdparm
        except ImportError:
            return self.rdparm_slow(fname)

        # The optimized parser only works on local, uncompressed files
        # TODO: Add gzip and bzip2 support to the optimized reader
        if (hasattr(fname, 'read') or slow
            or fname.startswith('http://') or fname.startswith('https://')
            or fname.startswith('ftp://')
            or fname.endswith('.bz2') or fname.endswith('.gz')):

            return self.rdparm_slow(fname)

        # We have the optimized version and a local file
        try:
            ret = _rdparm.rdparm(fname)
        except TypeError:
            # This is raised if VERSION is not found
            with closing(genopen(fname, 'r')) as f:
                return self.rdparm_old(f.readlines())
        else:
            # Unpack returned contents
            parm_data, parm_comments, formats, unkflg, flag_list, version = ret
            # Now assign them to instance attributes and process where necessary
            self.parm_data = parm_data
            self.parm_comments = parm_comments
            for key in formats:
                self.formats[key] = FortranFormat(formats[key])
            self.flag_list = flag_list
            self.version = version
            # Now we have to process all of those sections that the optimized
            # parser couldn't figure out
            for flag in unkflg:
                rawdata = self.parm_data[flag]
                self.parm_data[flag] = []
                for line in rawdata:
                    self.parm_data[flag].extend(self.formats[flag].read(line))
            if 'CTITLE' in self.parm_data:
                CHARGE_SCALE = CHARMM_ELECTROSTATIC
            else:
                CHARGE_SCALE = AMBER_ELECTROSTATIC
            try:
                for i, chg in enumerate(self.parm_data[self.charge_flag]):
                    self.parm_data[self.charge_flag][i] = chg / CHARGE_SCALE
            except KeyError:
                pass

    #===================================================

    def rdparm_slow(self, fname):
        """
        Parses the Amber format file. This parser is written in pure Python and
        is therefore slower than the C++-optimized version
        """

        current_flag = ''
        fmtre = re.compile(r'%FORMAT *\((.+)\)')
        version = None

        if isinstance(fname, string_types):
            prm = genopen(fname, 'r')
            own_handle = True
        elif hasattr(fname, 'read'):
            prm = fname
            own_handle = False
        else:
            raise TypeError('%s must be a file name or file-like object' % fname)

        # Open up the file and read the data into memory
        for line in prm:
            if line[0] == '%':
                if line[0:8] == '%VERSION':
                    self.version = line.strip()
                    continue
                elif line[0:5] == '%FLAG':
                    current_flag = line[6:].strip()
                    self.formats[current_flag] = ''
                    self.parm_data[current_flag] = []
                    self.parm_comments[current_flag] = []
                    self.flag_list.append(current_flag)
                    continue
                elif line[0:8] == '%COMMENT':
                    self.parm_comments[current_flag].append(line[9:].strip())
                    continue
                elif line[0:7] == '%FORMAT':
                    fmt = FortranFormat(fmtre.match(line).groups()[0])
                    # RESIDUE_ICODE can have a lot of blank data...
                    if current_flag == 'RESIDUE_ICODE':
                        fmt.read = fmt._read_nostrip
                    self.formats[current_flag] = fmt
                    continue
            try:
                self.parm_data[current_flag].extend(fmt.read(line))
            except KeyError:
                if version is not None:
                    raise
                break # Skip out of the loop down to the old-format parser

        # convert charges to fraction-electrons
        if 'CTITLE' in self.parm_data:
            CHARGE_SCALE = CHARMM_ELECTROSTATIC
        else:
            CHARGE_SCALE = AMBER_ELECTROSTATIC
        if self.charge_flag in self.parm_data:
            for i, chg in enumerate(self.parm_data[self.charge_flag]):
                self.parm_data[self.charge_flag][i] = chg / CHARGE_SCALE
        # If we don't have a version, then read in an old-file topology
        if self.version is None:
            prm.seek(0)
            return self.rdparm_old(prm.readlines())
        if own_handle:
            prm.close()
        return

    #===================================================

    def rdparm_old(self, prmtop_lines, check=False):
        """
        This reads an old-style topology file and stores the results in the
        same data structures as a new-style topology file

        Parameters
        ----------
        prmtop_lines : list of str
            List of all lines in the prmtop file
        check : bool, optional
            If True, only the first couple sections will be read to determine if
            this is, in fact, an old-style topology file
        """
        def read_integer(line_idx, lines, num_items):
            # line_idx should be the line _before_ the first line you
            # want data from.
            i, tmp_data = 0, []
            while i < num_items:
                idx = i % 12
                if idx == 0:
                    line_idx += 1
                try:
                    tmp_data.append(int(lines[line_idx][idx*6:idx*6+6]))
                except ValueError:
                    raise ValueError(
                            'Error parsing line %d, token %d [%s]: Problem '
                            'during integer read.' % (line_idx, idx,
                            lines[line_idx][idx*6:idx*6+6])
                    )
                i += 1
            # If we had no items, we need to jump a line:
            if num_items == 0: line_idx += 1
            return tmp_data, line_idx

        def read_string(line_idx, lines, num_items):
            # line_idx should be the line _before_ the first line you
            # want data from.
            i, tmp_data = 0, []
            while i < num_items:
                idx = i % 20
                if idx == 0:
                    line_idx += 1
                tmp_data.append(lines[line_idx][idx*4:idx*4+4])
                i += 1
            # If we had no items, we need to jump a line:
            if num_items == 0: line_idx += 1
            return tmp_data, line_idx

        def read_float(line_idx, lines, num_items):
            # line_idx should be the line _before_ the first line you
            # want data from.
            i, tmp_data = 0, []
            while i < num_items:
                idx = i % 5
                if idx == 0:
                    line_idx += 1
                try:
                    tmp_data.append(float(lines[line_idx][idx*16:idx*16+16]))
                except ValueError:
                    raise ValueError(
                            'Error parsing line %d, token %d [%s]: Problem '
                            'during floating point read.' % (line_idx, idx,
                            lines[line_idx][idx*16:idx*16+16])
                    )
                i += 1
            # If we had no items, we need to jump a line:
            if num_items == 0: line_idx += 1
            return tmp_data, line_idx

        # First add a title
        self.add_flag('TITLE', '20a4', data=['| Converted old-style topology'])

        # Next, read in the pointers
        line_idx = 0
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, 30)
        # Add a final pointer of 0, which corresponds to NUMEXTRA
        tmp_data.append(0)
        self.add_flag('POINTERS', '10I8', data=tmp_data)

        # Set some of the pointers we need
        natom = self.parm_data['POINTERS'][NATOM]
        ntypes = self.parm_data['POINTERS'][NTYPES]
        nres = self.parm_data['POINTERS'][NRES]
        numbnd = self.parm_data['POINTERS'][NUMBND]
        numang = self.parm_data['POINTERS'][NUMANG]
        nptra = self.parm_data['POINTERS'][NPTRA]
        natyp = self.parm_data['POINTERS'][NATYP]
        nbonh = self.parm_data['POINTERS'][NBONH]
        nbona = self.parm_data['POINTERS'][NBONA]
        ntheth = self.parm_data['POINTERS'][NTHETH]
        ntheta = self.parm_data['POINTERS'][NTHETA]
        nex = self.parm_data['POINTERS'][NEXT]
        nphia = self.parm_data['POINTERS'][NPHIA]
        nphb = self.parm_data['POINTERS'][NPHB]
        nphih = self.parm_data['POINTERS'][NPHIH]

        # This is enough to convince me that we have an old-style prmtop if we
        # have the number of integers I suspect we should
        if check:
            return len(tmp_data) == 31

        # Next read in the atom names
        tmp_data, line_idx = read_string(line_idx, prmtop_lines, natom)
        self.add_flag('ATOM_NAME', '20a4', data=tmp_data)

        # Next read the charges
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, natom)
        # Divide by the electrostatic constant
        tmp_data = [x / AMBER_ELECTROSTATIC for x in tmp_data]
        self.add_flag('CHARGE', '5E16.8', data=tmp_data)

        # Next read the masses
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, natom)
        self.add_flag('MASS', '5E16.8', data=tmp_data)

        # Next read atom type index
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, natom)
        self.add_flag('ATOM_TYPE_INDEX', '10I8', data=tmp_data)

        # Next read number excluded atoms
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, natom)
        self.add_flag('NUMBER_EXCLUDED_ATOMS', '10I8', data=tmp_data)

        # Next read nonbonded parm index
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, ntypes**2)
        self.add_flag('NONBONDED_PARM_INDEX', '10I8', data=tmp_data)

        # Next read residue label
        tmp_data, line_idx = read_string(line_idx, prmtop_lines, nres)
        self.add_flag('RESIDUE_LABEL', '20a4', data=tmp_data)

        # Next read residue pointer
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, nres)
        self.add_flag('RESIDUE_POINTER', '10I8', data=tmp_data)

        # Next read bond force constant
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, numbnd)
        self.add_flag('BOND_FORCE_CONSTANT', '5E16.8', data=tmp_data)

        # Next read bond equil value
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, numbnd)
        self.add_flag('BOND_EQUIL_VALUE', '5E16.8', data=tmp_data)

        # Next read angle force constant
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, numang)
        self.add_flag('ANGLE_FORCE_CONSTANT', '5E16.8', data=tmp_data)

        # Next read the angle equilibrium value
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, numang)
        self.add_flag('ANGLE_EQUIL_VALUE', '5E16.8', data=tmp_data)

        # Next read the dihedral force constant
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, nptra)
        self.add_flag('DIHEDRAL_FORCE_CONSTANT', '5E16.8', data=tmp_data)

        # Next read dihedral periodicity
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, nptra)
        self.add_flag('DIHEDRAL_PERIODICITY', '5E16.8', data=tmp_data)

        # Next read the dihedral phase
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, nptra)
        self.add_flag('DIHEDRAL_PHASE', '5E16.8', data=tmp_data)

        # Next read SOLTY (?)
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, natyp)
        self.add_flag('SOLTY', '5E16.8', data=tmp_data)

        # Next read lennard jones acoef and bcoef
        numvals = ntypes * (ntypes + 1) / 2
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, numvals)
        self.add_flag('LENNARD_JONES_ACOEF', '5E16.8', data=tmp_data)
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, numvals)
        self.add_flag('LENNARD_JONES_BCOEF', '5E16.8', data=tmp_data)

        # Next read bonds including hydrogen
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, nbonh*3)
        self.add_flag('BONDS_INC_HYDROGEN', '10I8', data=tmp_data)

        # Next read bonds without hydrogen
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, nbona*3)
        self.add_flag('BONDS_WITHOUT_HYDROGEN', '10I8', data=tmp_data)

        # Next read angles including hydrogen
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, ntheth*4)
        self.add_flag('ANGLES_INC_HYDROGEN', '10I8', data=tmp_data)

        # Next read angles without hydrogen
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, ntheta*4)
        self.add_flag('ANGLES_WITHOUT_HYDROGEN', '10I8', data=tmp_data)

        # Next read dihdrals including hydrogen
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, nphih*5)
        self.add_flag('DIHEDRALS_INC_HYDROGEN', '10I8', data=tmp_data)

        # Next read dihedrals without hydrogen
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, nphia*5)
        self.add_flag('DIHEDRALS_WITHOUT_HYDROGEN', '10I8', data=tmp_data)

        # Next read the excluded atoms list
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, nex)
        self.add_flag('EXCLUDED_ATOMS_LIST', '10I8', data=tmp_data)

        # Next read the hbond terms
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, nphb)
        self.add_flag('HBOND_ACOEF', '5E16.8', data=tmp_data)
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, nphb)
        self.add_flag('HBOND_BCOEF', '5E16.8', data=tmp_data)
        tmp_data, line_idx = read_float(line_idx, prmtop_lines, nphb)
        self.add_flag('HBCUT', '5E16.8', data=tmp_data)

        # Next read amber atom type
        tmp_data, line_idx = read_string(line_idx, prmtop_lines, natom)
        self.add_flag('AMBER_ATOM_TYPE', '20a4', data=tmp_data)

        # Next read tree chain classification
        tmp_data, line_idx = read_string(line_idx, prmtop_lines, natom)
        self.add_flag('TREE_CHAIN_CLASSIFICATION', '20a4', data=tmp_data)

        # Next read the join array
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, natom)
        self.add_flag('JOIN_ARRAY', '10I8', data=tmp_data)

        # Next read the irotat array
        tmp_data, line_idx = read_integer(line_idx, prmtop_lines, natom)
        self.add_flag('IROTAT', '10I8', data=tmp_data)

        # Now do PBC stuff
        if self.parm_data['POINTERS'][IFBOX]:
            # Solvent pointers
            tmp_data, line_idx = read_integer(line_idx, prmtop_lines, 3)
            self.add_flag('SOLVENT_POINTERS', '10I8', data=tmp_data)
            nspm = tmp_data[1]

            # Atoms per molecule
            tmp_data, line_idx = read_integer(line_idx, prmtop_lines, nspm)
            self.add_flag('ATOMS_PER_MOLECULE', '10I8', data=tmp_data)

            # Box dimensions
            tmp_data, line_idx = read_float(line_idx, prmtop_lines, 4)
            self.add_flag('BOX_DIMENSIONS', '5E16.8', data=tmp_data)

        # Now do CAP stuff
        if self.parm_data['POINTERS'][IFCAP]:
            # CAP_INFO
            tmp_data, line_idx = read_integer(line_idx, prmtop_lines, 1)
            self.add_flag('CAP_INFO', '10I8', data=tmp_data)
            tmp_data, line_idx = read_integer(line_idx, prmtop_lines, 4)
            self.add_flag('CAP_INFO2', '10I8', data=tmp_data)
        # end if self.parm_data['POINTERS'][IFCAP]

    #===================================================

    def set_version(self):
        """ Sets the version string """
        now = datetime.datetime.now()
        self.version = (
                '%%VERSION  VERSION_STAMP = V0001.000  DATE = %02d/%02d/%02d  '
                '%02d:%02d:%02d' % (now.month, now.day, now.year % 100,
                                    now.hour, now.minute, now.second)
        )

    #===================================================

    def write_parm(self, name):
        """
        Writes the current data in parm_data into a new topology file with
        the given name

        Parameters
        ----------
        name : str or file-like
            Name of the file to write the topology file to or file-like object to write
        """
        # now that we know we will write the new prmtop file, open the new file
        if isinstance(name, string_types):
            new_prm = genopen(name, 'w')
            own_handle = True
        else:
            new_prm = name
            own_handle = False
        try:
            # get current time to put into new prmtop file if we had a %VERSION
            self.set_version()
            # convert charges back to amber charges...
            if 'CTITLE' in self.parm_data:
                CHARGE_SCALE = CHARMM_ELECTROSTATIC
            else:
                CHARGE_SCALE = AMBER_ELECTROSTATIC

            if self.charge_flag in self.parm_data.keys():
                for i in range(len(self.parm_data[self.charge_flag])):
                    self.parm_data[self.charge_flag][i] *= CHARGE_SCALE
            # write version to top of prmtop file
            new_prm.write('%s\n' % self.version)

            # write data to prmtop file, inserting blank line if it's an empty field
            for flag in self.flag_list:
                new_prm.write('%%FLAG %s\n' % flag)
                # Insert any comments before the %FORMAT specifier
                for comment in self.parm_comments[flag]:
                    new_prm.write('%%COMMENT %s\n' % comment)
                new_prm.write('%%FORMAT(%s)\n' % self.formats[flag])
                if len(self.parm_data[flag]) == 0: # empty field...
                    new_prm.write('\n')
                    continue
                self.formats[flag].write(self.parm_data[flag], new_prm)
        finally:
            if own_handle:
                new_prm.close()

        if self.charge_flag in self.parm_data.keys():
            # Convert charges back to electron-units
            for i in range(len(self.parm_data[self.charge_flag])):
                self.parm_data[self.charge_flag][i] /= CHARGE_SCALE

    #===================================================

    def add_flag(self, flag_name, flag_format, data=None, num_items=-1,
                 comments=None, after=None):
        """
        Adds a new flag with the given flag name and Fortran format string and
        initializes the array with the values given, or as an array of 0s
        of length num_items

        Parameters
        ----------
        flag_name : str
            Name of the flag to insert. It is converted to all upper case
        flag_format : str
            Fortran format string representing how the data in this section
            should be written and read. Do not enclose in ()
        data : list=None
            Sequence with data for the new flag. If None, a list of zeros of
            length ``num_items`` (see below) is given as a holder
        num_items : int=-1
            Number of items in the section. This variable is ignored if a set of
            data are given in `data`
        comments : list of str=None
            List of comments to add to this section
        after : str=None
            If provided, the added flag will be added after the one with the
            name given to `after`. If this flag does not exist, IndexError will
            be raised

        Raises
        ------
        AmberError if flag already exists
        IndexError if the ``after`` flag does not exist
        """
        if flag_name in self.parm_data:
            raise AmberError('%s already exists' % (flag_name))
        if after is not None:
            after = after.upper()
            if not after in self.flag_list:
                raise IndexError('%s not found in topology flag list' % after)
            # If the 'after' flag is the last one, just append
            if self.flag_list[-1] == after:
                self.flag_list.append(flag_name.upper())
            else:
                # Otherwise find the index and add it after
                idx = self.flag_list.index(after) + 1
                self.flag_list.insert(idx, flag_name.upper())
        else:
            self.flag_list.append(flag_name.upper())
        self.formats[flag_name.upper()] = FortranFormat(flag_format)
        if data is not None:
            self.parm_data[flag_name.upper()] = list(data)
        else:
            if num_items < 0:
                raise AmberError("If you do not supply prmtop data, num_items "
                                 "must be non-negative!")
            self.parm_data[flag_name.upper()] = [0 for i in range(num_items)]
        if comments is not None:
            if isinstance(comments, string_types):
                comments = [comments]
            else:
                comments = list(comments)
            self.parm_comments[flag_name.upper()] = comments
        else:
            self.parm_comments[flag_name.upper()] = []

    #===================================================

    def delete_flag(self, flag_name):
        """ Removes a flag from the topology file """
        flag_name = flag_name.upper()
        if flag_name in self.flag_list:
            del self.flag_list[self.flag_list.index(flag_name)]
        if flag_name in self.parm_comments:
            del self.parm_comments[flag_name]
        if flag_name in self.formats:
            del self.formats[flag_name]
        if flag_name in self.parm_data:
            del self.parm_data[flag_name]

    #===================================================

    def __getstate__(self):
        return dict(parm_data=self.parm_data, flag_list=self.flag_list,
                    formats=self.formats, parm_comments=self.parm_comments,
                    charge_flag=self.charge_flag, version=self.version,
                    name=self.name)

    def __setstate__(self, d):
        self.__dict__ = d



'''
Miscellaneous functions, classes etc
'''
class has_next_iterator(object):
    '''
    A wrapper class for iterators so that the .has_next() method is implemented

    See - http://stackoverflow.com/questions/1966591/hasnext-in-python-iterators
    '''
    def __init__(self, it):
        self.it = iter(it)
        self._has_next = None
    def __iter__(self):
        return self
    def __next__(self):
        if self._has_next:
            result = self._the_next
        else:
            result = next(self.it)
        self._has_next = None
        return result
    def next(self):
        if self._has_next:
            result = self._the_next
        else:
            result = next(self.it)
        self._has_next = None
        return result
    def has_next(self):
        if self._has_next is None:
            try: 
                self._the_next = next(self.it)
            except StopIteration:
                self._has_next = False
            else:
                self._has_next = True
        return self._has_next


def expand_edit_descriptors(eds):
    expanded_eds = []
    for ed in eds:
        if hasattr(ed, 'repeat') and (ed.repeat is not None):
            expanded_eds.extend(ed.repeat * [ed])
        else:
            expanded_eds.append(ed)
    return expanded_eds



"""
This module contains all of the pb namelist variables for the amber
programs and automatically loads those dictionaries with the default
values found in that program (sander or sander.APBS)

                           GPL LICENSE INFO

Copyright (C) 2009 - 2011 Dwight Mcgee, Bill Miller, and Jason Swails

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
"""

class pb(object):

    def __init__(self):
        self.sander = {'epsin' : 1.0, 'epsout' : 80.0, 'smoothopt' : 1,
                       'istrng' : 0.0, 'pbtemp' : 300.0, 'radiopt' : 1,
                       'dprob' : 1.4, 'iprob' : 2.0, 'npbopt' : 0, 'solvopt' : 1,
                       'accept' : 0.001, 'maxitn' : 100, 'fillratio' : 2.0,
                       'space' : 0.5, 'nbuffer' : 0, 'nfocus' : 2, 'fscale' : 8,
                       'npbgrid' : 1, 'arcres' : 0.25, 'bcopt' : 5,
                       'scalec' : 0, 'eneopt' : 2, 'frcopt' : 0, 'cutfd' : 5.0,
                       'cutnb' : 0.0, 'nsnbr' : 1, 'nsnba' : 1, 'phiout' : 0,
                       'phiform' : 0, 'npbverb' : 0, 'npopt' : 2, 'decompopt' : 2,
                       'use_rmin' : 1, 'sprob' : 0.557, 'vprob' : 1.3,
                       'rhow_effect' : 1.129, 'use_sav' : 1,
                       'cavity_surften' : 0.0378, 'cavity_offset' : -0.5692,
                       'maxsph' : 400, 'maxarc' : 512, 'cutsa' : 9.0, 'ndofd' : 1,
                       'ndosas' : 1, 'fmiccg' : 0.9, 'ivalence' : 1.0,
                       'laccept' : 0.1, 'wsor' : 1.9, 'lwsor' : 1.95,
                       'pbkappa' : 0, 'radinc' : 0.8, 'expthresh' : 0.2,
                       'offx' : 0.0, 'offy' : 0.0, 'offz' : 0.0,
                       'sepbuf' : 4.0, 'mpopt' : 0, 'lmax' : 80, 'maxarcdot' : 1500,
                       'epsmem': 1.0, 'membraneopt' : 0, 'sasopt' : 0, 'mthick' : 40.0,
                       'mctrdz' : 0.0, 'mprob' : 2.7, 'poretype': 1, 'npbverb' : 0}

        self.sanderAPBS = {'dime' : '0,0,0', 'pdime' : '0,0,0', 'cglen' : '0,0,0',
                           'fglen' : '0,0,0', 'grid' : '0,0,0', 'nlev' : 4,
                           'nonlin' : 0, 'bcfl' : 1, 'nion' : 0, 'pdie' : 2.0,
                           'sdie' : 78.4, 'srfm' : 2, 'chgm' : 1, 'temp' : 298.15,
                           'gamma' : 0.105, 'sdens' : 10.0, 'calc_type' : 1,
                           'calcnpenergy' : 1, 'cmeth' : 1, 'ccmeth' : 1,
                           'fcmeth' : 1, 'ionq' : 'NO_DEFAULT',
                           'ionc' : 'NO_DEFAULT', 'ionrr' : 'NO_DEFAULT',
                           'smvolume' : 10, 'smsize' : 1000.0, 'srad' : 1.4,
                           'swin' : 0.3, 'calcenergy' : 2, 'calcforce' : 2,
                           'apbs_debug' : 0, 'sp_apbs' : '.false.',
                           'apbs_print' : 1, 'wpot' : 0, 'wchg' : 0, 'wsmol' : 0,
                           'ispara' : 0, 'radiopt' : 0, 'geom_upd_limit' : 0,
                           'evdw_upd_limit' : 0, 'pqr' : 0, 'dime_updates' : 0,
                           'wkappa' : 0, 'wdiel' : 0, 'rchg' : 0, 'rkappa' : 0,
                           'rdiel' : 0, 'calcnpforce' : 2, 'watompot' : 0,
                           'rpot' : 0}



"""
This package contains classes responsible for loading and dumping PyRosetta
Pose objects.
"""

from __future__ import print_function

from parmed.exceptions import RosettaError
from parmed.periodic_table import AtomicNum, Mass
from parmed.structure import Structure
from parmed.topologyobjects import Atom, ExtraPoint, Bond
from parmed.utils.six.moves import range

try:
    from pyrosetta import Pose, AtomID
except ImportError:
    Pose = AtomID = None

def _n_prior(pose, nbr):
    prior = -1
    for i in range(1, nbr.rsd()):
        prior += pose.residue(i).natoms()
    return prior + nbr.atomno()

class RosettaPose(object):

    @staticmethod
    def load(pose):
        """
        Load a :class:`Pose` object and return a populated :class:`Structure`
        instance

        Parameters
        ----------
        pose : :class:`Pose`
            PyRosetta :class:`Pose` object to convert
        """
        if not Pose or not AtomID:
            raise ImportError('Could not load the PyRosetta module.')
        if not isinstance(pose, Pose):
            raise TypeError('Object is not a PyRosetta Pose object.')

        struct = Structure()

        atnum = 1
        conf = pose.conformation()
        for resid in range(1, pose.total_residue()+1):
            res = pose.residue(resid)
            resname = res.name3().strip()
            chain = chr(res.chain()+ord('A')-1)
            for atno, at in enumerate(res.atoms(), start=1):
                try:
                    atinfo = res.atom_type(atno)
                    atname = res.atom_name(atno).strip()
                    if atinfo.is_virtual():
                        atsym = 'EP'
                    else:
                        atsym = atinfo.element()
                    rmin = atinfo.lj_radius()
                    epsilon = atinfo.lj_wdepth()
                    atomic_number = AtomicNum[atsym]
                    mass = Mass[atsym]
                except KeyError:
                    raise RosettaError('Could not recognize element: %s.'
                                       % atsym)

                params = dict(atomic_number=atomic_number, name=atname,
                              charge=0.0, mass=mass, occupancy=0.0,
                              bfactor=0.0, altloc='', number=atnum,
                              rmin=rmin, epsilon=epsilon)

                if atinfo.is_virtual():
                    atom = ExtraPoint(**params)
                else:
                    atom = Atom(**params)

                atom.xx, atom.xy, atom.xz = (at.xyz()[0], at.xyz()[1], at.xyz()[2])

                struct.add_atom(atom, resname, resid, chain, '')
                atnum += 1
                try:
                    for nbr in conf.bonded_neighbor_all_res(AtomID(atno,
                                                                   resid)):
                        if nbr.rsd() < resid or (nbr.rsd() == resid
                                                 and nbr.atomno() < atno):
                            struct.bonds.append(
                                Bond(struct.atoms[_n_prior(pose, nbr)],
                                     atom))
                except:
                    raise RosettaError('Could not add bonds.')

        struct.unchange()
        return struct



" This command adds a new Lennard Jones atom type from the selected atoms. "

from math import sqrt
from parmed.constants import NTYPES
from parmed.utils.six.moves import range

def AddLJType(parm, sel_atms, radius, epsilon, radius14, epsilon14):
    """ Adds a new Lennard Jones type to a topology file """
    # Set the new atom type
    old_type = None
    for i in range(len(sel_atms)):
        if sel_atms[i] == 1:
            if old_type is None:
                old_type = parm.parm_data['ATOM_TYPE_INDEX'][i]
            parm.parm_data['ATOM_TYPE_INDEX'][i] = parm.ptr('ntypes') + 1

    # Now increment NTYPES
    parm.parm_data['POINTERS'][NTYPES] += 1
    parm.pointers['NTYPES'] += 1
   
    # Now create a whole new array for NONBONDED_PARM_INDEX
    start_idx = max(parm.parm_data['NONBONDED_PARM_INDEX']) + 1
    current_idx = max(parm.parm_data['NONBONDED_PARM_INDEX']) + 1
    old_ntypes = parm.ptr('ntypes') - 1
    for i in range(old_ntypes):
        # Copy over the first ntypes terms
        parm.parm_data['NONBONDED_PARM_INDEX'].insert(
                                    parm.ptr('ntypes')*(i+1)-1, current_idx)
        current_idx += 1

    # Now add the interaction of the last type with every other
    for i in range(old_ntypes):
        parm.parm_data['NONBONDED_PARM_INDEX'].append(start_idx)
        start_idx += 1

    # New type interacting with itself
    parm.parm_data['NONBONDED_PARM_INDEX'].append(current_idx)

    # Now we need to add onto the ACOEF and BCOEF arrays
    for i in range(old_ntypes):
        rad = parm.LJ_radius[i] + radius
        depth = sqrt(parm.LJ_depth[i] * epsilon)
        parm.parm_data['LENNARD_JONES_ACOEF'].append(depth * rad**12)
        parm.parm_data['LENNARD_JONES_BCOEF'].append(2 * depth * rad**6)
    # Do the same for the 1-4 interactions if we're doing a chamber prmtop
    if parm.chamber:
        for i in range(old_ntypes):
            rad = parm.LJ_14_radius[i] + radius14
            depth = sqrt(parm.LJ_14_depth[i] * epsilon14)
            parm.parm_data['LENNARD_JONES_14_ACOEF'].append(depth * rad**12)
            parm.parm_data['LENNARD_JONES_14_BCOEF'].append(2 * depth * rad**6)
   
    # Now add the last type interacting with itself, and add it to the
    # LJ_radius/depth arrays
    depth = epsilon
    rad = 2 * radius
    parm.parm_data['LENNARD_JONES_ACOEF'].append(depth * rad**12)
    parm.parm_data['LENNARD_JONES_BCOEF'].append(2 * depth * rad**6)
    parm.LJ_radius.append(radius)
    parm.LJ_depth.append(depth)

    # Do the same for chamber prmtops
    if parm.chamber:
        depth = epsilon14
        rad = 2 * radius14
        parm.parm_data['LENNARD_JONES_14_ACOEF'].append(depth * rad**12)
        parm.parm_data['LENNARD_JONES_14_BCOEF'].append(2 * depth * rad**6)
        parm.LJ_14_radius.append(radius14)
        parm.LJ_14_depth.append(epsilon14)

    # If our prmtop had a C-coefficient array, just copy the terms from the old
    # atom type to the new atom type, as defined by the type of the first atom
    # that got assigned the "new" type.
    if 'LENNARD_JONES_CCOEF' in parm.parm_data:
        ccoeffs = parm.parm_data['LENNARD_JONES_CCOEF']
        for i in range(old_ntypes):
            nbi = parm.ptr('ntypes')*(old_type-1) + i
            idx = parm.parm_data['NONBONDED_PARM_INDEX'][nbi] - 1
            ccoeffs.append(ccoeffs[idx])

        # Now the last type interacting with itself
        nbi = parm.ptr('ntypes')*(old_type-1) + old_type - 1
        idx = parm.parm_data['NONBONDED_PARM_INDEX'][nbi] - 1
        ccoeffs.append(ccoeffs[idx])
